"""Musical transcription query system.

This module provides a comprehensive query system for analyzing musical 
transcriptions, with exact compatibility to the TypeScript implementation.
"""

from __future__ import annotations
from typing import List, Optional, Dict, Union, Any, Tuple
import json

from .classes.piece import Piece
from .classes.trajectory import Trajectory
from .classes.pitch import Pitch
from .enums import Instrument
from .query_types import (
    CategoryType, DesignatorType, SegmentationType, QueryType,
    QueryAnswerType, MultipleReturnType, MultipleOptionType
)
from .sequence_utils import find_sequence_indexes, loose_sequence_indexes, split_trajs_by_silences


class Query:
    """Query system for musical transcription analysis.
    
    This class provides sophisticated filtering and analysis capabilities
    for musical transcriptions, maintaining exact compatibility with the
    TypeScript implementation.
    """
    
    def __init__(self, piece: Piece, options: Optional[Dict[str, Any]] = None):
        """Initialize a new Query instance.
        
        Args:
            piece: The Piece object to query
            options: Query configuration options
        """
        opts = options or {}
        
        # Core properties
        self.piece = piece
        self.trajectories: List[List[Trajectory]] = []
        self.phrase_idxs: List[int] = []
        self.instrument_idx: int = opts.get("instrument_idx", 0)
        self.identifier: List[Union[int, str, Dict[str, int]]] = []
        self.stringified_identifier: List[str] = []
        self.sequence_length: Optional[int] = opts.get("sequence_length")
        self.repetition: bool = False
        self.start_times: List[float] = []
        
        # Query parameters
        self.designator = DesignatorType(opts.get("designator", "includes"))
        self.category = CategoryType(opts.get("category", "trajectoryID"))
        self.segmentation = SegmentationType(opts.get("segmentation", "phrase"))
        self.max_dur: float = opts.get("max_dur", 60.0)
        self.min_dur: float = opts.get("min_dur", 0.0)
        
        # Category-specific parameters
        self.consonant: Optional[str] = opts.get("consonant")
        self.vowel: Optional[str] = opts.get("vowel")
        self.trajectory_id: Optional[int] = opts.get("trajectory_id")
        self.pitch: Optional[Pitch] = opts.get("pitch")
        self.pitch_sequence: Optional[List[Pitch]] = opts.get("pitch_sequence")
        self.traj_id_sequence: Optional[List[int]] = opts.get("traj_id_sequence")
        
        # Section categorization parameters
        self.section_top_level: Optional[str] = opts.get("section_top_level")
        self.alap_section: Optional[str] = opts.get("alap_section")
        self.comp_type: Optional[str] = opts.get("comp_type")
        self.comp_sec_tempo: Optional[str] = opts.get("comp_sec_tempo")
        self.tala: Optional[str] = opts.get("tala")
        
        # Phrase categorization parameters
        self.phrase_type: Optional[str] = opts.get("phrase_type")
        self.elaboration_type: Optional[str] = opts.get("elaboration_type")
        self.vocal_art_type: Optional[str] = opts.get("vocal_art_type")
        self.inst_art_type: Optional[str] = opts.get("inst_art_type")
        self.incidental: Optional[str] = opts.get("incidental")
        
        # Validate parameters and execute query
        self._validate_parameters()
        self._execute_filters()
        self._filter_by_duration()
        self._generate_identifiers()
        self._calculate_start_times()
    
    def _validate_parameters(self) -> None:
        """Validate query parameters based on category and other constraints."""
        # Instrument index validation
        if self.instrument_idx < 0 or self.instrument_idx >= len(self.piece.instrumentation):
            raise ValueError(f"instrument_idx {self.instrument_idx} is out of range. Piece has {len(self.piece.instrumentation)} instruments (valid indices: 0-{len(self.piece.instrumentation)-1})")
        
        # Sequence length validation
        if self.segmentation == SegmentationType.SEQUENCE_OF_TRAJECTORIES:
            if self.sequence_length is None:
                raise ValueError("sequence_length is required when segmentation is sequenceOfTrajectories")
        
        # Category-specific validation
        consonant_categories = {
            CategoryType.STARTING_CONSONANT, 
            CategoryType.ENDING_CONSONANT, 
            CategoryType.ANY_CONSONANT
        }
        
        section_query = False
        
        if self.category == CategoryType.PITCH:
            if self.pitch is None:
                raise ValueError("pitch is required when category is pitch")
        elif self.category == CategoryType.TRAJECTORY_ID:
            if self.trajectory_id is None:
                raise ValueError("trajectory_id is required when category is trajectoryID")
        elif self.category == CategoryType.VOWEL:
            inst = self.piece.instrumentation[self.instrument_idx]
            if inst not in [Instrument.Vocal_F, Instrument.Vocal_M]:
                raise ValueError("category vowel is only for vocal instruments")
            if self.vowel is None:
                raise ValueError("vowel is required when category is vowel")
        elif self.category in consonant_categories:
            inst = self.piece.instrumentation[self.instrument_idx]
            if inst not in [Instrument.Vocal_F, Instrument.Vocal_M]:
                raise ValueError("category consonant is only for vocal instruments")
            if self.consonant is None:
                raise ValueError("consonant is required when category is consonant")
        elif self.category in [CategoryType.PITCH_SEQUENCE_STRICT, CategoryType.PITCH_SEQUENCE_LOOSE]:
            if self.pitch_sequence is None:
                raise ValueError("pitch_sequence is required")
        elif self.category in [CategoryType.TRAJ_SEQUENCE_STRICT, CategoryType.TRAJ_SEQUENCE_LOOSE]:
            if self.traj_id_sequence is None:
                raise ValueError("traj_id_sequence is required")
        elif self.category == CategoryType.SECTION_TOP_LEVEL:
            if self.section_top_level is None:
                raise ValueError("section_top_level is required")
            section_query = True
        elif self.category == CategoryType.ALAP_SECTION:
            if self.alap_section is None:
                raise ValueError("alap_section is required")
            section_query = True
        elif self.category == CategoryType.COMP_TYPE:
            if self.comp_type is None:
                raise ValueError("comp_type is required")
            section_query = True
        elif self.category == CategoryType.COMP_SEC_TEMPO:
            if self.comp_sec_tempo is None:
                raise ValueError("comp_sec_tempo is required")
            section_query = True
        elif self.category == CategoryType.TALA:
            if self.tala is None:
                raise ValueError("tala is required")
            section_query = True
        
        # Section query constraints
        if section_query:
            if self.designator in [DesignatorType.ENDS_WITH, DesignatorType.STARTS_WITH]:
                raise ValueError("Section queries cannot be used with startsWith or endsWith")
            if self.segmentation in [
                SegmentationType.SEQUENCE_OF_TRAJECTORIES,
                SegmentationType.CONNECTED_SEQUENCE_OF_TRAJECTORIES
            ]:
                raise ValueError("Section queries cannot be used with trajectory sequence segmentation")
    
    def _execute_filters(self) -> None:
        """Execute the appropriate filter based on segmentation type."""
        if self.segmentation == SegmentationType.PHRASE:
            self._phrase_filter()
        elif self.segmentation == SegmentationType.GROUP:
            self._group_filter()
        elif self.segmentation == SegmentationType.SEQUENCE_OF_TRAJECTORIES:
            self._sequence_of_trajectories_filter()
        elif self.segmentation == SegmentationType.CONNECTED_SEQUENCE_OF_TRAJECTORIES:
            self._connected_sequence_of_trajectories_filter()
    
    def _phrase_filter(self) -> None:
        """Filter phrases based on query criteria."""
        phrases = self.piece.phrase_grid[self.instrument_idx]
        filtered_phrases = []
        
        for phrase in phrases:
            if self._test_phrase_match(phrase):
                filtered_phrases.append(phrase)
        
        self.trajectories = [phrase.trajectories for phrase in filtered_phrases]
        self.identifier = [phrase.piece_idx for phrase in filtered_phrases if phrase.piece_idx is not None]
    
    def _group_filter(self) -> None:
        """Filter groups based on query criteria."""
        groups = self.piece.all_groups({"instrument_idx": self.instrument_idx})
        filtered_groups = []
        
        for group in groups:
            if self._test_group_match(group):
                filtered_groups.append(group)
        
        self.trajectories = [group.trajectories for group in filtered_groups]
        for group in filtered_groups:
            self.identifier.append(group.id)
    
    def _sequence_of_trajectories_filter(self) -> None:
        """Filter sequences of trajectories based on query criteria."""
        all_trajs = self.piece.all_trajectories(self.instrument_idx)
        
        for i in range(len(all_trajs) - self.sequence_length + 1):
            traj_seq = all_trajs[i:i + self.sequence_length]
            
            if self._test_trajectory_sequence_match(traj_seq):
                self.trajectories.append(traj_seq)
                id_obj = {
                    "phrase_idx": traj_seq[0].phrase_idx,
                    "traj_idx": traj_seq[0].num
                }
                self.identifier.append(id_obj)
    
    def _connected_sequence_of_trajectories_filter(self) -> None:
        """Filter connected sequences of trajectories (split by silences)."""
        all_trajs = self.piece.all_trajectories(self.instrument_idx)
        split_trajs = split_trajs_by_silences(all_trajs)
        
        for traj_seq in split_trajs:
            if self._test_trajectory_sequence_match(traj_seq):
                self.trajectories.append(traj_seq)
                id_obj = {
                    "phrase_idx": traj_seq[0].phrase_idx,
                    "traj_idx": traj_seq[0].num
                }
                self.identifier.append(id_obj)
    
    def _test_phrase_match(self, phrase) -> bool:
        """Test if a phrase matches the query criteria."""
        if self.category == CategoryType.PITCH:
            trial_arr = [pitch.numbered_pitch for pitch in phrase.all_pitches(self.repetition)]
            return self._pitch_diff(self.pitch, trial_arr)
        elif self.category == CategoryType.TRAJECTORY_ID:
            return self._traj_id_diff(self.trajectory_id, phrase.trajectories)
        elif self.category == CategoryType.VOWEL:
            trial_arr = [traj.vowel for traj in phrase.trajectories]
            return self._vowel_diff(self.vowel, trial_arr)
        elif self.category == CategoryType.STARTING_CONSONANT:
            trial_arr = [traj.start_consonant for traj in phrase.trajectories]
            return self._consonant_diff(self.consonant, trial_arr)
        elif self.category == CategoryType.ENDING_CONSONANT:
            trial_arr = [traj.end_consonant for traj in phrase.trajectories]
            return self._consonant_diff(self.consonant, trial_arr)
        elif self.category == CategoryType.ANY_CONSONANT:
            trial_arr = []
            for traj in phrase.trajectories:
                trial_arr.extend([traj.start_consonant, traj.end_consonant])
            return self._consonant_diff(self.consonant, trial_arr)
        elif self.category == CategoryType.PITCH_SEQUENCE_STRICT:
            trial_arr = [pitch.numbered_pitch for pitch in phrase.all_pitches(self.repetition)]
            return self._pitch_seq_strict_diff(self.pitch_sequence, trial_arr)
        elif self.category == CategoryType.PITCH_SEQUENCE_LOOSE:
            trial_arr = [pitch.numbered_pitch for pitch in phrase.all_pitches(self.repetition)]
            return self._pitch_seq_loose_diff(self.pitch_sequence, trial_arr)
        elif self.category == CategoryType.TRAJ_SEQUENCE_STRICT:
            trial_arr = [traj.id for traj in phrase.trajectories]
            return self._traj_seq_strict_diff(self.traj_id_sequence, trial_arr)
        elif self.category == CategoryType.TRAJ_SEQUENCE_LOOSE:
            trial_arr = [traj.id for traj in phrase.trajectories]
            return self._traj_seq_loose_diff(self.traj_id_sequence, trial_arr)
        elif self.category == CategoryType.SECTION_TOP_LEVEL:
            p_idx = phrase.piece_idx
            return self._sec_top_level_diff(p_idx)
        elif self.category == CategoryType.ALAP_SECTION:
            p_idx = phrase.piece_idx
            return self._alap_section_diff(p_idx)
        elif self.category == CategoryType.COMP_TYPE:
            p_idx = phrase.piece_idx
            return self._comp_type_diff(p_idx)
        elif self.category == CategoryType.COMP_SEC_TEMPO:
            p_idx = phrase.piece_idx
            return self._comp_sec_tempo_diff(p_idx)
        elif self.category == CategoryType.TALA:
            p_idx = phrase.piece_idx
            return self._tala_diff(p_idx)
        elif self.category in [
            CategoryType.PHRASE_TYPE,
            CategoryType.ELABORATION_TYPE,
            CategoryType.VOCAL_ART_TYPE,
            CategoryType.INST_ART_TYPE,
            CategoryType.INCIDENTAL
        ]:
            p_idx = phrase.piece_idx
            return self._phrase_label_diff(p_idx)
        
        return False
    
    def _test_group_match(self, group) -> bool:
        """Test if a group matches the query criteria."""
        # Similar logic to phrase match but for groups
        # This would need to be implemented based on the group structure
        # For now, return False as placeholder
        return False
    
    def _test_trajectory_sequence_match(self, traj_seq: List[Trajectory]) -> bool:
        """Test if a trajectory sequence matches the query criteria."""
        if self.category == CategoryType.PITCH:
            pitches = []
            for traj in traj_seq:
                pitches.extend(traj.pitches)
            n_pitches = [pitch.numbered_pitch for pitch in pitches]
            if not self.repetition:
                n_pitches = self._remove_consecutive_duplicates(n_pitches)
            return self._pitch_diff(self.pitch, n_pitches)
        elif self.category == CategoryType.TRAJECTORY_ID:
            return self._traj_id_diff(self.trajectory_id, traj_seq)
        elif self.category == CategoryType.VOWEL:
            vowels = [traj.vowel for traj in traj_seq]
            return self._vowel_diff(self.vowel, vowels)
        elif self.category == CategoryType.STARTING_CONSONANT:
            consonants = [traj.start_consonant for traj in traj_seq]
            return self._consonant_diff(self.consonant, consonants)
        elif self.category == CategoryType.ENDING_CONSONANT:
            consonants = [traj.end_consonant for traj in traj_seq]
            return self._consonant_diff(self.consonant, consonants)
        elif self.category == CategoryType.ANY_CONSONANT:
            consonants = []
            for traj in traj_seq:
                consonants.extend([traj.start_consonant, traj.end_consonant])
            return self._consonant_diff(self.consonant, consonants)
        elif self.category == CategoryType.PITCH_SEQUENCE_STRICT:
            pitches = []
            for traj in traj_seq:
                pitches.extend(traj.pitches)
            n_pitches = [pitch.numbered_pitch for pitch in pitches]
            if not self.repetition:
                n_pitches = self._remove_consecutive_duplicates(n_pitches)
            return self._pitch_seq_strict_diff(self.pitch_sequence, n_pitches)
        elif self.category == CategoryType.PITCH_SEQUENCE_LOOSE:
            pitches = []
            for traj in traj_seq:
                pitches.extend(traj.pitches)
            n_pitches = [pitch.numbered_pitch for pitch in pitches]
            if not self.repetition:
                n_pitches = self._remove_consecutive_duplicates(n_pitches)
            return self._pitch_seq_loose_diff(self.pitch_sequence, n_pitches)
        elif self.category == CategoryType.TRAJ_SEQUENCE_STRICT:
            traj_ids = [traj.id for traj in traj_seq]
            return self._traj_seq_strict_diff(self.traj_id_sequence, traj_ids)
        elif self.category == CategoryType.TRAJ_SEQUENCE_LOOSE:
            traj_ids = [traj.id for traj in traj_seq]
            return self._traj_seq_loose_diff(self.traj_id_sequence, traj_ids)
        elif self.category == CategoryType.SECTION_TOP_LEVEL:
            raise ValueError("section_top_level cannot be used with trajectory sequence segmentation")
        
        return False
    
    @staticmethod
    def _remove_consecutive_duplicates(sequence: List[int]) -> List[int]:
        """Remove consecutive duplicate values from a sequence."""
        if not sequence:
            return []
        
        result = [sequence[0]]
        for i in range(1, len(sequence)):
            if sequence[i] != sequence[i-1]:
                result.append(sequence[i])
        return result
    
    # ---- Diff methods for different query types ----
    
    def _pitch_diff(self, pitch: Pitch, n_pitches: List[int]) -> bool:
        """Test pitch criteria against a list of numbered pitches."""
        boolean = False
        
        if self.designator == DesignatorType.INCLUDES:
            boolean = pitch.numbered_pitch in n_pitches
        elif self.designator == DesignatorType.EXCLUDES:
            boolean = pitch.numbered_pitch not in n_pitches
        elif self.designator == DesignatorType.STARTS_WITH:
            boolean = len(n_pitches) > 0 and n_pitches[0] == pitch.numbered_pitch
        elif self.designator == DesignatorType.ENDS_WITH:
            boolean = len(n_pitches) > 0 and n_pitches[-1] == pitch.numbered_pitch
        
        return boolean
    
    def _pitch_seq_strict_diff(self, pitch_seq: List[Pitch], n_pitches: List[int]) -> bool:
        """Test strict pitch sequence criteria."""
        boolean = False
        num_pitch_seq = [pitch.numbered_pitch for pitch in pitch_seq]
        
        if self.designator == DesignatorType.INCLUDES:
            boolean = len(find_sequence_indexes(num_pitch_seq, n_pitches)) > 0
        elif self.designator == DesignatorType.EXCLUDES:
            boolean = len(find_sequence_indexes(num_pitch_seq, n_pitches)) == 0
        elif self.designator == DesignatorType.STARTS_WITH:
            if len(num_pitch_seq) <= len(n_pitches):
                boolean = all(num_pitch_seq[i] == n_pitches[i] for i in range(len(num_pitch_seq)))
        elif self.designator == DesignatorType.ENDS_WITH:
            if len(num_pitch_seq) <= len(n_pitches):
                start_idx = len(n_pitches) - len(num_pitch_seq)
                boolean = all(num_pitch_seq[i] == n_pitches[start_idx + i] for i in range(len(num_pitch_seq)))
        
        return boolean
    
    def _pitch_seq_loose_diff(self, pitch_seq: List[Pitch], n_pitches: List[int]) -> bool:
        """Test loose pitch sequence criteria."""
        boolean = False
        num_pitch_seq = [pitch.numbered_pitch for pitch in pitch_seq]
        
        if self.designator == DesignatorType.INCLUDES:
            loose_obj = loose_sequence_indexes(num_pitch_seq, n_pitches)
            boolean = loose_obj["truth"]
        elif self.designator == DesignatorType.EXCLUDES:
            loose_obj = loose_sequence_indexes(num_pitch_seq, n_pitches)
            boolean = not loose_obj["truth"]
        elif self.designator == DesignatorType.STARTS_WITH:
            loose_obj = loose_sequence_indexes(num_pitch_seq, n_pitches)
            boolean = loose_obj["truth"] and loose_obj["first_idx"] == 0
        elif self.designator == DesignatorType.ENDS_WITH:
            loose_obj = loose_sequence_indexes(num_pitch_seq, n_pitches)
            boolean = loose_obj["truth"] and loose_obj["last_idx"] == len(n_pitches) - 1
        
        return boolean
    
    def _traj_seq_strict_diff(self, traj_id_seq: List[int], full_traj_list: List[int]) -> bool:
        """Test strict trajectory sequence criteria."""
        boolean = False
        
        if self.designator == DesignatorType.INCLUDES:
            boolean = len(find_sequence_indexes(traj_id_seq, full_traj_list)) > 0
        elif self.designator == DesignatorType.EXCLUDES:
            boolean = len(find_sequence_indexes(traj_id_seq, full_traj_list)) == 0
        elif self.designator == DesignatorType.STARTS_WITH:
            indexes = find_sequence_indexes(traj_id_seq, full_traj_list)
            boolean = len(indexes) > 0 and indexes[0] == 0
        elif self.designator == DesignatorType.ENDS_WITH:
            start_idx = len(full_traj_list) - len(traj_id_seq)
            indexes = find_sequence_indexes(traj_id_seq, full_traj_list)
            boolean = len(indexes) > 0 and indexes[0] == start_idx
        
        return boolean
    
    def _traj_seq_loose_diff(self, traj_id_seq: List[int], full_traj_list: List[int]) -> bool:
        """Test loose trajectory sequence criteria."""
        boolean = False
        
        if self.designator == DesignatorType.INCLUDES:
            loose_obj = loose_sequence_indexes(traj_id_seq, full_traj_list)
            boolean = loose_obj["truth"]
        elif self.designator == DesignatorType.EXCLUDES:
            loose_obj = loose_sequence_indexes(traj_id_seq, full_traj_list)
            boolean = not loose_obj["truth"]
        elif self.designator == DesignatorType.STARTS_WITH:
            loose_obj = loose_sequence_indexes(traj_id_seq, full_traj_list)
            boolean = loose_obj["truth"] and loose_obj["first_idx"] == 0
        elif self.designator == DesignatorType.ENDS_WITH:
            loose_obj = loose_sequence_indexes(traj_id_seq, full_traj_list)
            boolean = loose_obj["truth"] and loose_obj["last_idx"] == len(full_traj_list) - 1
        
        return boolean
    
    def _traj_id_diff(self, traj_id: int, trajectories: List[Trajectory]) -> bool:
        """Test trajectory ID criteria against a list of trajectories."""
        boolean = False
        
        if self.designator == DesignatorType.INCLUDES:
            boolean = any(traj.id == traj_id for traj in trajectories)
        elif self.designator == DesignatorType.EXCLUDES:
            boolean = not any(traj.id == traj_id for traj in trajectories)
        elif self.designator == DesignatorType.STARTS_WITH:
            boolean = len(trajectories) > 0 and trajectories[0].id == traj_id
        elif self.designator == DesignatorType.ENDS_WITH:
            boolean = len(trajectories) > 0 and trajectories[-1].id == traj_id
        
        return boolean
    
    def _vowel_diff(self, vowel: str, vowels: List[Optional[str]]) -> bool:
        """Test vowel criteria against a list of vowels."""
        boolean = False
        
        if self.designator == DesignatorType.INCLUDES:
            boolean = vowel in vowels
        elif self.designator == DesignatorType.EXCLUDES:
            boolean = vowel not in vowels
        elif self.designator == DesignatorType.STARTS_WITH:
            boolean = len(vowels) > 0 and vowels[0] == vowel
        elif self.designator == DesignatorType.ENDS_WITH:
            boolean = len(vowels) > 0 and vowels[-1] == vowel
        
        return boolean
    
    def _consonant_diff(self, consonant: str, consonants: List[Optional[str]]) -> bool:
        """Test consonant criteria against a list of consonants."""
        boolean = False
        
        if self.designator == DesignatorType.INCLUDES:
            boolean = consonant in consonants
        elif self.designator == DesignatorType.EXCLUDES:
            boolean = consonant not in consonants
        elif self.designator == DesignatorType.STARTS_WITH:
            boolean = len(consonants) > 0 and consonants[0] == consonant
        # Note: endsWith is not implemented for consonants in TypeScript version
        
        return boolean
    
    def _sec_top_level_diff(self, p_idx: int) -> bool:
        """Test section top level criteria."""
        boolean = False
        s_idx = self.piece.s_idx_from_p_idx(p_idx, self.instrument_idx)
        if s_idx is not None and self.piece.sections:
            section = self.piece.sections[s_idx]
            if self.designator == DesignatorType.INCLUDES:
                boolean = section.categorization.get("Top Level") == self.section_top_level
            elif self.designator == DesignatorType.EXCLUDES:
                boolean = section.categorization.get("Top Level") != self.section_top_level
        
        return boolean
    
    def _alap_section_diff(self, p_idx: int) -> bool:
        """Test alap section criteria."""
        boolean = False
        s_idx = self.piece.s_idx_from_p_idx(p_idx)
        if s_idx is not None and self.piece.sections:
            section = self.piece.sections[s_idx]
            alap_cat = section.categorization.get("Alap", {})
            if self.designator == DesignatorType.INCLUDES:
                boolean = alap_cat.get(self.alap_section, False)
            elif self.designator == DesignatorType.EXCLUDES:
                boolean = not alap_cat.get(self.alap_section, False)
        
        return boolean
    
    def _comp_type_diff(self, p_idx: int) -> bool:
        """Test composition type criteria."""
        boolean = False
        s_idx = self.piece.s_idx_from_p_idx(p_idx)
        if s_idx is not None and self.piece.sections:
            section = self.piece.sections[s_idx]
            comp_cat = section.categorization.get("Composition Type", {})
            if self.designator == DesignatorType.INCLUDES:
                boolean = comp_cat.get(self.comp_type, False)
            elif self.designator == DesignatorType.EXCLUDES:
                boolean = not comp_cat.get(self.comp_type, False)
        
        return boolean
    
    def _comp_sec_tempo_diff(self, p_idx: int) -> bool:
        """Test composition section/tempo criteria."""
        boolean = False
        s_idx = self.piece.s_idx_from_p_idx(p_idx)
        if s_idx is not None and self.piece.sections:
            section = self.piece.sections[s_idx]
            comp_sec_tempo_cat = section.categorization.get("Comp.-section/Tempo", {})
            if self.designator == DesignatorType.INCLUDES:
                boolean = comp_sec_tempo_cat.get(self.comp_sec_tempo, False)
            elif self.designator == DesignatorType.EXCLUDES:
                boolean = not comp_sec_tempo_cat.get(self.comp_sec_tempo, False)
        
        return boolean
    
    def _tala_diff(self, p_idx: int) -> bool:
        """Test tala criteria."""
        boolean = False
        s_idx = self.piece.s_idx_from_p_idx(p_idx)
        if s_idx is not None and self.piece.sections:
            section = self.piece.sections[s_idx]
            tala_cat = section.categorization.get("Tala", {})
            if self.designator == DesignatorType.INCLUDES:
                boolean = tala_cat.get(self.tala, False)
            elif self.designator == DesignatorType.EXCLUDES:
                boolean = not tala_cat.get(self.tala, False)
        
        return boolean
    
    def _phrase_label_diff(self, p_idx: int) -> bool:
        """Test phrase label criteria."""
        boolean = False
        
        if p_idx < len(self.piece.phrases):
            phrase = self.piece.phrases[p_idx]
            if hasattr(phrase, 'categorization_grid') and phrase.categorization_grid:
                cat = phrase.categorization_grid[0]
                
                if self.category == CategoryType.PHRASE_TYPE:
                    boolean = cat.get("Phrase", {}).get(self.phrase_type, False)
                elif self.category == CategoryType.ELABORATION_TYPE:
                    boolean = cat.get("Elaboration", {}).get(self.elaboration_type, False)
                elif self.category == CategoryType.VOCAL_ART_TYPE:
                    boolean = cat.get("Vocal Articulation", {}).get(self.vocal_art_type, False)
                elif self.category == CategoryType.INST_ART_TYPE:
                    boolean = cat.get("Instrumental Articulation", {}).get(self.inst_art_type, False)
                elif self.category == CategoryType.INCIDENTAL:
                    boolean = cat.get("Incidental", {}).get(self.incidental, False)
                
                if self.designator == DesignatorType.EXCLUDES:
                    boolean = not boolean
        
        return boolean
    
    # ---- Helper methods ----
    
    def _filter_by_duration(self) -> None:
        """Filter results by duration constraints."""
        remove_idxs = []
        
        for t_idx, traj_seq in enumerate(self.trajectories):
            if not traj_seq:
                continue
                
            first_phrase_idx = traj_seq[0].phrase_idx
            if first_phrase_idx is None or first_phrase_idx >= len(self.piece.phrases):
                continue
                
            first_phrase = self.piece.phrases[first_phrase_idx]
            start = traj_seq[0].start_time + (first_phrase.start_time or 0)
            
            last_phrase_idx = traj_seq[-1].phrase_idx
            if last_phrase_idx is None or last_phrase_idx >= len(self.piece.phrases):
                continue
                
            last_phrase = self.piece.phrases[last_phrase_idx]
            end = traj_seq[-1].end_time + (last_phrase.start_time or 0)
            
            duration = end - start
            
            if duration > self.max_dur or duration < self.min_dur:
                remove_idxs.append(t_idx)
        
        # Remove filtered items
        self.trajectories = [
            traj_seq for t_idx, traj_seq in enumerate(self.trajectories) 
            if t_idx not in remove_idxs
        ]
        self.identifier = [
            identifier for t_idx, identifier in enumerate(self.identifier) 
            if t_idx not in remove_idxs
        ]
    
    def _generate_identifiers(self) -> None:
        """Generate stringified identifiers for results."""
        self.stringified_identifier = [json.dumps(identifier) for identifier in self.identifier]
    
    def _calculate_start_times(self) -> None:
        """Calculate start times for all trajectory sequences."""
        self.start_times = []
        
        for traj_seq in self.trajectories:
            if not traj_seq:
                self.start_times.append(0.0)
                continue
                
            phrase_idx = traj_seq[0].phrase_idx
            if phrase_idx is None or phrase_idx >= len(self.piece.phrase_grid[self.instrument_idx]):
                self.start_times.append(0.0)
                continue
                
            phrase = self.piece.phrase_grid[self.instrument_idx][phrase_idx]
            start_time = (traj_seq[0].start_time or 0) + (phrase.start_time or 0)
            self.start_times.append(start_time)
    
    @property
    def query_answers(self) -> List[QueryAnswerType]:
        """Get structured query results."""
        answers = []
        
        for t_idx, trajs in enumerate(self.trajectories):
            if not trajs:
                continue
                
            # Calculate timing
            start_phrase_idx = trajs[0].phrase_idx
            if start_phrase_idx is None or start_phrase_idx >= len(self.piece.phrase_grid[self.instrument_idx]):
                continue
                
            start_phrase = self.piece.phrase_grid[self.instrument_idx][start_phrase_idx]
            start_time = (trajs[0].start_time or 0) + (start_phrase.start_time or 0)
            
            end_phrase_idx = trajs[-1].phrase_idx
            if end_phrase_idx is None or end_phrase_idx >= len(self.piece.phrase_grid[self.instrument_idx]):
                continue
                
            end_phrase = self.piece.phrase_grid[self.instrument_idx][end_phrase_idx]
            end_time = (trajs[-1].end_time or 0) + (end_phrase.start_time or 0)
            
            duration = end_time - start_time
            
            # Generate title
            title = self._generate_title(trajs, start_phrase_idx, end_phrase_idx)
            
            answer = QueryAnswerType(
                trajectories=trajs,
                identifier=self.identifier[t_idx],
                title=title,
                start_time=start_time,
                end_time=end_time,
                duration=duration,
                segmentation=self.segmentation
            )
            answers.append(answer)
        
        return answers
    
    def _generate_title(self, trajs: List[Trajectory], start_phrase_idx: int, end_phrase_idx: int) -> str:
        """Generate a descriptive title for the query result."""
        if self.segmentation == SegmentationType.PHRASE:
            phrase_idx_str = str(start_phrase_idx + 1)
            return f"Phrase {phrase_idx_str}"
        elif self.segmentation == SegmentationType.GROUP:
            # This would need group information from trajectories
            phrase_idx_str = str(start_phrase_idx + 1)
            group_id = getattr(trajs[0], 'group_id', None)
            return f"Phrase {phrase_idx_str} Group {group_id if group_id else 'Unknown'}"
        elif self.segmentation == SegmentationType.SEQUENCE_OF_TRAJECTORIES:
            return self._generate_trajectory_sequence_title(trajs, start_phrase_idx, end_phrase_idx)
        elif self.segmentation == SegmentationType.CONNECTED_SEQUENCE_OF_TRAJECTORIES:
            return self._generate_trajectory_sequence_title(trajs, start_phrase_idx, end_phrase_idx)
        
        return "Unknown"
    
    def _generate_trajectory_sequence_title(self, trajs: List[Trajectory], start_phrase_idx: int, end_phrase_idx: int) -> str:
        """Generate title for trajectory sequence results."""
        phrase_idxs = list(set(traj.phrase_idx for traj in trajs if traj.phrase_idx is not None))
        phrase_idxs.sort()
        
        first_traj_idx = (trajs[0].num or 0) + 1
        last_traj_idx = (trajs[-1].num or 0) + 1
        
        if len(phrase_idxs) == 1:
            phrase_idx_str = str(phrase_idxs[0] + 1)
            return f"Phrase {phrase_idx_str} Traj {first_traj_idx}-{last_traj_idx}"
        else:
            start_phrase_str = str(phrase_idxs[0] + 1)
            end_phrase_str = str(phrase_idxs[-1] + 1)
            return f"Phrase {start_phrase_str} Traj {first_traj_idx} - Phrase {end_phrase_str} Traj {last_traj_idx}"
    
    # ---- Static factory methods ----
    
    @staticmethod
    async def single(
        transcription_id: str = "63445d13dc8b9023a09747a6",
        segmentation: SegmentationType = SegmentationType.PHRASE,
        designator: DesignatorType = DesignatorType.INCLUDES,
        category: CategoryType = CategoryType.TRAJECTORY_ID,
        pitch: Optional[Pitch] = None,
        sequence_length: Optional[int] = None,
        trajectory_id: Optional[int] = None,
        vowel: Optional[str] = None,
        consonant: Optional[str] = None,
        instrument_idx: int = 0,
    ) -> Query:
        """Create a single query instance.
        
        This is an async method that would need to fetch the piece data.
        For now, it's a placeholder that would need integration with the API client.
        """
        # This would need to be implemented with actual piece fetching logic
        # For now, return a basic implementation
        raise NotImplementedError("Single query requires integration with SwaraClient")
    
    @staticmethod
    async def multiple(
        queries: List[QueryType],
        transcription_id: str = "63445d13dc8b9023a09747a6",
        piece: Optional[Piece] = None,
        segmentation: SegmentationType = SegmentationType.PHRASE,
        sequence_length: Optional[int] = None,
        min_dur: float = 0.0,
        max_dur: float = 60.0,
        every: bool = True,
        instrument_idx: int = 0,
    ) -> MultipleReturnType:
        """Execute multiple queries and combine results.
        
        Args:
            queries: List of query specifications
            transcription_id: ID of transcription to query
            piece: Pre-loaded Piece object (optional)
            segmentation: Segmentation type for all queries
            sequence_length: Sequence length for trajectory sequences
            min_dur: Minimum duration filter
            max_dur: Maximum duration filter
            every: If True, require all queries to match; if False, any query can match
            instrument_idx: Index of instrument track to query
            
        Returns:
            Tuple of (trajectories, identifiers, query_answers)
        """
        if not queries:
            raise ValueError("No queries provided")
        
        if piece is None:
            # This would need integration with SwaraClient to fetch the piece
            raise NotImplementedError("Multiple queries require piece data or SwaraClient integration")
        
        output_trajectories: List[List[Trajectory]] = []
        output_identifiers: List[str] = []
        query_answers: List[QueryAnswerType] = []
        non_stringified_output_identifiers: List[Union[int, str, Dict[str, int]]] = []
        
        # Create query objects
        query_objs = []
        for query in queries:
            query_options = {
                "segmentation": segmentation,
                "designator": query.get("designator"),
                "category": query.get("category"),
                "pitch": query.get("pitch"),
                "sequence_length": sequence_length,
                "trajectory_id": query.get("trajectory_id"),
                "vowel": query.get("vowel"),
                "consonant": query.get("consonant"),
                "pitch_sequence": query.get("pitch_sequence"),
                "traj_id_sequence": query.get("traj_id_sequence"),
                "section_top_level": query.get("section_top_level"),
                "alap_section": query.get("alap_section"),
                "comp_type": query.get("comp_type"),
                "comp_sec_tempo": query.get("comp_sec_tempo"),
                "tala": query.get("tala"),
                "phrase_type": query.get("phrase_type"),
                "elaboration_type": query.get("elaboration_type"),
                "vocal_art_type": query.get("vocal_art_type"),
                "inst_art_type": query.get("inst_art_type"),
                "incidental": query.get("incidental"),
                "min_dur": min_dur,
                "max_dur": max_dur,
                "instrument_idx": instrument_idx,
            }
            query_objs.append(Query(piece, query_options))
        
        if every:
            # Only select trajectories that are in all answers
            if query_objs:
                output_identifiers = query_objs[0].stringified_identifier[:]
                for answer in query_objs[1:]:
                    output_identifiers = [
                        id_str for id_str in output_identifiers 
                        if id_str in answer.stringified_identifier
                    ]
                
                # Get corresponding trajectories and answers
                idxs = [
                    query_objs[0].stringified_identifier.index(id_str) 
                    for id_str in output_identifiers
                ]
                output_trajectories = [query_objs[0].trajectories[idx] for idx in idxs]
                non_stringified_output_identifiers = [query_objs[0].identifier[idx] for idx in idxs]
                query_answers = [query_objs[0].query_answers[idx] for idx in idxs]
        else:
            # Select trajectories that are in any answer
            start_times = []
            seen_ids = set()
            
            for answer in query_objs:
                for s_idx, s_id in enumerate(answer.stringified_identifier):
                    if s_id not in seen_ids:
                        seen_ids.add(s_id)
                        output_identifiers.append(s_id)
                        output_trajectories.append(answer.trajectories[s_idx])
                        non_stringified_output_identifiers.append(answer.identifier[s_idx])
                        query_answers.append(answer.query_answers[s_idx])
                        start_times.append(answer.start_times[s_idx])
            
            # Sort by start times
            sort_idxs = sorted(range(len(start_times)), key=lambda i: start_times[i])
            output_trajectories = [output_trajectories[idx] for idx in sort_idxs]
            non_stringified_output_identifiers = [non_stringified_output_identifiers[idx] for idx in sort_idxs]
            query_answers = [query_answers[idx] for idx in sort_idxs]
        
        return output_trajectories, non_stringified_output_identifiers, query_answers


import os
import sys
import pytest
from unittest.mock import patch

sys.path.insert(0, os.path.abspath('.'))

from idtap.classes.phrase import Phrase, init_phrase_categorization
from idtap.classes.trajectory import Trajectory
from idtap.classes.pitch import Pitch
from idtap.classes.articulation import Articulation
from idtap.classes.group import Group
from idtap.classes.raga import Raga
from idtap.classes.chikari import Chikari


def make_phrase(trajectories, start_time=None):
    return Phrase({'trajectories': trajectories, 'start_time': start_time})

# ------------------------------------------------------------------
# Tests from branch "codex/create-tests-for-getgroups-and-others"
# ------------------------------------------------------------------

def test_groups_retrieval():
    t1 = Trajectory()
    t2 = Trajectory()
    phrase = make_phrase([t1, t2])
    phrase.piece_idx = 0
    phrase.assign_phrase_idx()
    g = Group({'trajectories': [phrase.trajectories[0], phrase.trajectories[1]]})
    phrase.get_groups(0).append(g)
    assert phrase.get_groups(0) == [g]
    assert phrase.get_group_from_id(g.id) == g
    assert phrase.get_group_from_id('missing') is None
    with pytest.raises(Exception):
        phrase.get_groups(1)


def test_assign_phrase_idx_sets_indices():
    ts = [Trajectory(), Trajectory()]
    phrase = make_phrase(ts)
    phrase.piece_idx = 3
    phrase.assign_phrase_idx()
    for t in ts:
        assert t.phrase_idx == 3
    phrase.piece_idx = None
    phrase.assign_phrase_idx()
    for t in ts:
        assert t.phrase_idx is None


def test_assign_traj_nums():
    ts = [Trajectory(), Trajectory(), Trajectory()]
    phrase = make_phrase(ts)
    for t in ts:
        t.num = 99
    phrase.assign_traj_nums()
    for i, t in enumerate(ts):
        assert t.num == i
    empty = Phrase()
    empty.assign_traj_nums()  # should not raise


def test_assign_start_times():
    ts = [Trajectory(), Trajectory(), Trajectory()]
    phrase = make_phrase(ts)
    phrase.assign_start_times()
    expected = [0, 1, 2]
    for t, e in zip(ts, expected):
        assert pytest.approx(t.start_time, rel=1e-6) == e

    p2 = Phrase()
    p2.dur_array = None
    with pytest.raises(Exception):
        p2.assign_start_times()
    p3 = Phrase()
    p3.dur_tot = None
    p3.dur_array = [0.5, 0.5]
    with pytest.raises(Exception):
        p3.assign_start_times()


def test_update_fundamental():
    t1 = Trajectory({'pitches': [Pitch()]})
    t2 = Trajectory({'pitches': [Pitch()]})
    phrase = make_phrase([t1, t2])
    phrase.update_fundamental(440)
    for p in t1.pitches + t2.pitches:
        assert pytest.approx(p.fundamental, rel=1e-6) == 440


def test_all_pitches():
    p1 = Pitch({'swara': 'sa'})
    p2 = Pitch({'swara': 're'})
    t1 = Trajectory({'pitches': [p1]})
    t2 = Trajectory({'pitches': [p2]})
    t3 = Trajectory({'pitches': [Pitch({'swara': 're'})]})
    silent = Trajectory({'id': 12, 'pitches': [Pitch()]})
    phrase = make_phrase([t1, t2, t3, silent])
    assert phrase.all_pitches() == [p1, p2, t3.pitches[0]]
    assert phrase.all_pitches(False) == [p1, p2]


def test_swara_property():
    ts = [Trajectory(), Trajectory(), Trajectory()]
    phrase = make_phrase(ts, 10)
    times = [o['time'] for o in phrase.swara]
    assert times == [10, 11, 12]

    p2 = Phrase({'trajectories': [Trajectory()]})
    with pytest.raises(Exception):
        _ = p2.swara

    p3 = make_phrase([Trajectory()], 0)
    p3.trajectories[0].start_time = None
    with pytest.raises(Exception):
        _ = p3.swara

    p4 = make_phrase([Trajectory()], 0)
    p4.trajectories[0].dur_array = None
    with pytest.raises(Exception):
        _ = p4.swara

# ------------------------------------------------------------------
# Tests from branch "porting-project"
# ------------------------------------------------------------------

def test_phrase_methods_and_serialization():
    t1 = Trajectory({'num': 0, 'dur_tot': 0.5, 'pitches': [Pitch()]})
    t2 = Trajectory({'num': 1, 'dur_tot': 0.5, 'pitches': [Pitch({'swara': 'r'})]})
    p = Phrase({'trajectories': [t1, t2], 'raga': Raga()})
    assert pytest.approx(p.dur_tot, rel=1e-6) == 1
    assert pytest.approx(p.compute(0.25), rel=1e-6) == t1.compute(0.5)
    assert p.get_range()['min']['numberedPitch'] == t1.pitches[0].numbered_pitch
    nv = p.to_note_view_phrase()
    assert len(nv.pitches) == 2
    json_obj = p.to_json()
    copy = Phrase.from_json(json_obj)
    assert pytest.approx(copy.dur_tot, rel=1e-6) == 1
    assert len(copy.trajectories) == 2


def test_phrase_utility_functions():
    r = Raga()
    t1 = Trajectory({'num': 0, 'dur_tot': 0.5, 'pitches': [Pitch()]})
    t1.add_consonant('ka')
    t1.update_vowel('a')
    silent1 = Trajectory({'num': 1, 'id': 12, 'dur_tot': 0.25, 'pitches': [Pitch()]})
    silent2 = Trajectory({'num': 2, 'id': 12, 'dur_tot': 0.25, 'pitches': [Pitch()]})
    t2 = Trajectory({'num': 3, 'dur_tot': 0.5, 'pitches': [Pitch({'swara': 'r'})]})
    t2.update_vowel('i')
    p = Phrase({'trajectories': [t1, silent1, silent2, t2], 'raga': r, 'start_time': 0})
    p.reset()
    p.chikaris['0.3'] = Chikari({})

    idxs = p.first_traj_idxs()
    assert 0 in idxs
    assert 3 in idxs
    assert p.traj_idx_from_time(0.1) == 0

    chiks = p.chikaris_during_traj(t1, 0)
    assert len(chiks) == 1
    p.consolidate_silent_trajs()
    assert len(p.trajectories) == 3


def test_to_note_view_phrase_id0():
    pitch = Pitch({'swara': 'ga'})
    traj = Trajectory({'id': 0, 'pitches': [pitch], 'articulations': {'0.00': Articulation({'name': 'pluck'})}})
    phrase = Phrase({'trajectories': [traj]})
    nv = phrase.to_note_view_phrase()
    assert len(nv.pitches) == 1
    assert nv.pitches[0] == pitch


def test_to_note_view_phrase_nonzero():
    pitch = Pitch({'swara': 'ma'})
    traj = Trajectory({'id': 2, 'pitches': [pitch], 'articulations': {}})
    phrase = Phrase({'trajectories': [traj]})
    nv = phrase.to_note_view_phrase()
    assert pitch in nv.pitches


def test_from_json_reconstructs_grids():
    t1 = Trajectory({'num': 0, 'pitches': [Pitch()]})
    t2 = Trajectory({'num': 1, 'pitches': [Pitch({'swara': 'r'})]})
    c1 = Chikari({})
    obj = {
        'trajectoryGrid': [[t1.to_json(), t2.to_json()]],
        'chikariGrid': [{'0.5': c1.to_json()}],
        'instrumentation': ['Sitar', 'Violin'],
        'startTime': 0,
    }
    phrase = Phrase.from_json(obj)
    assert isinstance(phrase.trajectory_grid[0][0], Trajectory)
    assert isinstance(phrase.trajectory_grid[0][1], Trajectory)
    assert len(phrase.trajectory_grid) == 2
    assert isinstance(phrase.trajectory_grid[1], list)
    assert len(phrase.trajectory_grid[1]) == 0
    assert isinstance(phrase.chikari_grid[0]['0.5'], Chikari)
    assert len(phrase.chikari_grid) == 2
    assert len(phrase.chikari_grid[1].keys()) == 0


def test_traj_idx_from_time_error():
    t1 = Trajectory({'num': 0, 'dur_tot': 0.5, 'pitches': [Pitch()]})
    t2 = Trajectory({'num': 1, 'dur_tot': 0.5, 'pitches': [Pitch()]})
    phrase = Phrase({'trajectories': [t1, t2], 'start_time': 0})
    with pytest.raises(Exception):
        phrase.traj_idx_from_time(1.1)


def test_swara_edge_cases_shorter_durarray():
    t = Trajectory({'id': 1, 'pitches': [Pitch(), Pitch({'swara': 1})], 'dur_array': [1], 'dur_tot': 1})
    phrase = make_phrase([t], 0)
    sw = phrase.swara
    assert len(sw) == 1
    assert sw[0]['pitch'] == t.pitches[0]
    assert pytest.approx(sw[0]['time'], rel=1e-6) == 0


def test_swara_edge_cases_equal():
    t = Trajectory({'id': 7, 'pitches': [Pitch(), Pitch({'swara': 1})], 'dur_array': [0.4, 0.6], 'dur_tot': 1})
    phrase = make_phrase([t], 0)
    sw = phrase.swara
    assert len(sw) == 2
    assert pytest.approx(sw[0]['time'], rel=1e-6) == 0
    assert pytest.approx(sw[1]['time'], rel=1e-6) == 0.4


def test_consolidate_silent_trajs():
    t1 = Trajectory({'num': 0, 'dur_tot': 0.5, 'pitches': [Pitch()]})
    s1 = Trajectory({'num': 1, 'id': 12, 'dur_tot': 0.1, 'pitches': [Pitch()]})
    s2 = Trajectory({'num': 2, 'id': 12, 'dur_tot': 0.2, 'pitches': [Pitch()]})
    t2 = Trajectory({'num': 3, 'dur_tot': 0.5, 'pitches': [Pitch({'swara': 'r'})]})
    s3 = Trajectory({'num': 4, 'id': 12, 'dur_tot': 0.1, 'pitches': [Pitch()]})
    s4 = Trajectory({'num': 5, 'id': 12, 'dur_tot': 0.2, 'pitches': [Pitch()]})
    t3 = Trajectory({'num': 6, 'dur_tot': 0.5, 'pitches': [Pitch({'swara': 'g'})]})
    s5 = Trajectory({'num': 7, 'id': 12, 'dur_tot': 0.1, 'pitches': [Pitch()]})
    s6 = Trajectory({'num': 8, 'id': 12, 'dur_tot': 0.2, 'pitches': [Pitch()]})

    p = Phrase({'trajectories': [t1, s1, s2, t2, s3, s4, t3, s5, s6], 'raga': Raga(), 'start_time': 0})
    p.consolidate_silent_trajs()
    assert len(p.trajectories) == 6
    assert pytest.approx(p.trajectories[1].dur_tot, rel=1e-6) == 0.3
    assert pytest.approx(p.trajectories[3].dur_tot, rel=1e-6) == 0.3
    assert pytest.approx(p.trajectories[5].dur_tot, rel=1e-6) == 0.3


def test_consolidate_silent_trajs_missing_num():
    good = Trajectory({'num': 0, 'dur_tot': 0.5, 'pitches': [Pitch()]})
    bad = Trajectory({'id': 12, 'dur_tot': 0.5, 'pitches': [Pitch()]})
    p = Phrase({'trajectories': [good, bad], 'raga': Raga()})
    p.trajectories[1].num = None
    with pytest.raises(Exception):
        p.consolidate_silent_trajs()


def test_realign_pitches():
    t1 = Trajectory({'pitches': [Pitch(), Pitch({'swara': 'r'})]})
    t2 = Trajectory({'pitches': [Pitch({'swara': 'g'})]})
    r = Raga()
    phrase = Phrase({'trajectories': [t1, t2], 'raga': r})
    originals = [t.pitches[:] for t in phrase.trajectories]
    phrase.realign_pitches()
    for ti, traj in enumerate(phrase.trajectories):
        for pi, p in enumerate(traj.pitches):
            assert p is not originals[ti][pi]
            assert p.ratios == r.stratified_ratios


def test_compute_errors_and_null():
    p = Phrase()
    p.dur_array = None
    with pytest.raises(Exception):
        p.compute(0.5)

    p2 = Phrase()
    assert p2.dur_array == []
    assert p2.compute(0.5) is None


def test_durtot_and_durarray_preserved_with_empty_trajs():
    with patch.object(Phrase, 'dur_array_from_trajectories') as spy:
        p = Phrase({'dur_tot': 2, 'dur_array': [1]})
        assert p.dur_tot == 2
        assert p.dur_array == [1]
        assert not spy.called


def test_constructor_pads_grids():
    t1 = Trajectory()
    trajectory_grid = [[t1]]
    chikari_grid = [{}]
    instrumentation = ['Sitar', 'Violin']
    phrase = Phrase({'trajectory_grid': trajectory_grid, 'chikari_grid': chikari_grid, 'instrumentation': instrumentation})
    # Test that grids are padded correctly for multiple instruments
    assert len(phrase.trajectory_grid) == len(instrumentation)
    assert len(phrase.chikari_grid) == len(instrumentation)
    assert phrase.trajectory_grid[0] == [t1]  # Original content preserved
    assert phrase.trajectory_grid[1] == []    # Padded for second instrument
    assert phrase.chikari_grid[0] == {}       # Original content preserved  
    assert phrase.chikari_grid[1] == {}       # Padded for second instrument


def test_from_json_fallback_grids():
    obj = {'trajectories': [], 'instrumentation': ['Sitar', 'Violin'], 'startTime': 0}
    phrase = Phrase.from_json(obj)
    assert len(phrase.trajectory_grid) == len(obj['instrumentation'])
    assert len(phrase.chikari_grid) == len(obj['instrumentation'])
    for row in phrase.trajectory_grid:
        assert row == []
    for col in phrase.chikari_grid:
        assert col == {}


def test_chikaris_setter():
    phrase = Phrase()
    c = Chikari({})
    phrase.chikaris = {'0.1': c}
    assert phrase.chikari_grid[0]['0.1'] is c


def test_constructor_scales_trajectories():
    t1 = Trajectory({'dur_tot': 1, 'pitches': [Pitch()]})
    t2 = Trajectory({'dur_tot': 1, 'pitches': [Pitch({'swara': 'r'})]})
    dur_array = [0.2, 0.8]
    phrase = Phrase({'trajectories': [t1, t2], 'dur_tot': 4, 'dur_array': dur_array})
    assert pytest.approx(phrase.dur_tot, rel=1e-6) == 4
    assert phrase.dur_array == dur_array
    assert pytest.approx(t1.dur_tot, rel=1e-6) == 0.8
    assert pytest.approx(t2.dur_tot, rel=1e-6) == 3.2


def test_constructor_fills_missing_grids():
    traj = Trajectory()
    instrumentation = ['Sitar', 'Violin', 'Sarod']
    phrase = Phrase({'trajectories': [traj], 'instrumentation': instrumentation})
    assert len(phrase.trajectory_grid) == len(instrumentation)
    assert len(phrase.chikari_grid) == len(instrumentation)
    assert phrase.trajectory_grid[0] == [traj]
    assert phrase.chikari_grid[0] == {}
    assert phrase.trajectory_grid[1] == []
    assert phrase.trajectory_grid[2] == []
    assert phrase.chikari_grid[1] == {}
    assert phrase.chikari_grid[2] == {}


def test_missing_bol_alap_initialized():
    custom = init_phrase_categorization()
    del custom['Elaboration']['Bol Alap']
    phrase = Phrase({'categorization_grid': [custom]})
    assert phrase.categorization_grid[0]['Elaboration']['Bol Alap'] is False


# ----------------------------------------------------------------------
# is_section_start Tests (Issue #47)
# ----------------------------------------------------------------------

def test_is_section_start_true():
    """Test phrase with is_section_start = True."""
    phrase = Phrase({
        'trajectories': [],
        'is_section_start': True
    })
    assert phrase.is_section_start is True


def test_is_section_start_false():
    """Test phrase with is_section_start = False."""
    phrase = Phrase({
        'trajectories': [],
        'is_section_start': False
    })
    assert phrase.is_section_start is False


def test_is_section_start_none_default():
    """Test phrase without is_section_start (defaults to None)."""
    phrase = Phrase({
        'trajectories': []
    })
    assert phrase.is_section_start is None


def test_is_section_start_type_validation():
    """Test that non-boolean is_section_start raises TypeError."""
    with pytest.raises(TypeError, match="Parameter 'is_section_start' must be a boolean"):
        Phrase({
            'trajectories': [],
            'is_section_start': 'true'  # String instead of bool
        })

    with pytest.raises(TypeError, match="Parameter 'is_section_start' must be a boolean"):
        Phrase({
            'trajectories': [],
            'is_section_start': 1  # Integer instead of bool
        })


def test_is_section_start_serialization():
    """Test that is_section_start is included in serialization."""
    phrase_true = Phrase({
        'trajectories': [],
        'is_section_start': True
    })
    json_true = phrase_true.to_json()
    assert 'isSectionStart' in json_true
    assert json_true['isSectionStart'] is True

    phrase_false = Phrase({
        'trajectories': [],
        'is_section_start': False
    })
    json_false = phrase_false.to_json()
    assert 'isSectionStart' in json_false
    assert json_false['isSectionStart'] is False

    phrase_none = Phrase({
        'trajectories': []
    })
    json_none = phrase_none.to_json()
    assert 'isSectionStart' in json_none
    assert json_none['isSectionStart'] is None


def test_is_section_start_round_trip():
    """Test that is_section_start survives serialization and deserialization."""
    phrase = Phrase({
        'trajectories': [Trajectory({'dur_tot': 1})],
        'is_section_start': True,
        'raga': Raga()
    })

    json_obj = phrase.to_json()
    copy = Phrase.from_json(json_obj)

    assert copy.is_section_start is True
    assert copy.to_json()['isSectionStart'] == phrase.to_json()['isSectionStart']



from __future__ import annotations
from typing import List, Optional, Dict
import uuid

from .trajectory import Trajectory
from .pitch import Pitch

class Group:
    def __init__(self, options: Optional[Dict] = None) -> None:
        opts = options or {}
        
        # Parameter validation
        self._validate_parameters(opts)
        self.trajectories: List[Trajectory] = opts.get('trajectories', [])
        self.id: str = str(opts.get('id', uuid.uuid4()))
        # sort and validate
        self.trajectories.sort(key=lambda t: self._require_num(t))
        if len(self.trajectories) < 2:
            raise ValueError('Group must have at least 2 trajectories')
        if not self.test_for_adjacency():
            raise ValueError('Trajectories are not adjacent')
        for traj in self.trajectories:
            traj.group_id = self.id

    def _validate_parameters(self, opts: dict) -> None:
        """Validate constructor parameters and provide helpful error messages."""
        if not opts:
            return
            
        # Define allowed parameter names
        allowed_keys = {'trajectories', 'id'}
        provided_keys = set(opts.keys())
        invalid_keys = provided_keys - allowed_keys
        
        # Check for invalid parameter names
        if invalid_keys:
            error_messages = []
            
            for key in invalid_keys:
                if key == 'trajectory_list' or key == 'trajs':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'trajectories'?")
                elif key == 'group_id' or key == 'unique_id':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'id'?")
                else:
                    error_messages.append(f"Invalid parameter: '{key}'")
            
            error_msg = "; ".join(error_messages)
            error_msg += f". Allowed parameters: {sorted(allowed_keys)}"
            raise ValueError(error_msg)
        
        # Validate parameter types
        if 'trajectories' in opts:
            if not isinstance(opts['trajectories'], list):
                raise TypeError(f"Parameter 'trajectories' must be a list, got {type(opts['trajectories']).__name__}")
        
        if 'id' in opts and opts['id'] is not None:
            if not isinstance(opts['id'], str):
                # Allow UUID objects to be converted
                opts['id'] = str(opts['id'])

    def _require_num(self, traj: Trajectory) -> int:
        if traj.num is None:
            raise ValueError('Trajectory must have a num')
        return traj.num

    @property
    def min_freq(self) -> float:
        return min(t.min_freq for t in self.trajectories)

    @property
    def max_freq(self) -> float:
        return max(t.max_freq for t in self.trajectories)

    def all_pitches(self, repetition: bool = True) -> List[Pitch]:
        pitches: List[Pitch] = []
        for t in self.trajectories:
            if t.id != 12:
                pitches.extend(t.pitches)
        if not repetition:
            out: List[Pitch] = []
            for i, p in enumerate(pitches):
                if i == 0:
                    out.append(p)
                else:
                    prev = out[-1]
                    if not (p.swara == prev.swara and p.oct == prev.oct and p.raised == prev.raised):
                        out.append(p)
            return out
        return pitches

    def test_for_adjacency(self) -> bool:
        phrase_idxs = {t.phrase_idx for t in self.trajectories}
        if len(phrase_idxs) != 1:
            return False
        nums = [self._require_num(t) for t in self.trajectories]
        nums.sort()
        diffs = [nums[i+1]-nums[i] for i in range(len(nums)-1)]
        return all(d == 1 for d in diffs)

    def add_traj(self, traj: Trajectory) -> None:
        self.trajectories.append(traj)
        self.trajectories.sort(key=lambda t: self._require_num(t))
        if not self.test_for_adjacency():
            raise ValueError('Trajectories are not adjacent')
        traj.group_id = self.id

    def to_json(self) -> Dict:
        return {
            'trajectories': self.trajectories,
            'id': self.id,
        }

    @staticmethod
    def from_json(obj: Dict) -> 'Group':
        trajs = obj.get('trajectories', [])
        return Group({'trajectories': trajs, 'id': obj.get('id')})



import os
import sys
import math
import pytest

sys.path.insert(0, os.path.abspath('.'))

from idtap.classes.trajectory import Trajectory
from idtap.classes.pitch import Pitch
from idtap.classes.articulation import Articulation
from idtap.classes.automation import Automation
from idtap.classes.raga import Raga
from idtap.classes.phrase import Phrase
from idtap.classes.piece import Piece, durations_of_fixed_pitches
from idtap.enums import Instrument


def lin_space(start_val: float, stop_val: float, cardinality: int):
    step = (stop_val - start_val) / (cardinality - 1)
    return [start_val + step * i for i in range(cardinality)]


def test_default_trajectory():
    t = Trajectory()
    assert isinstance(t, Trajectory)
    assert t.id == 0
    assert t.pitches == [Pitch()]
    assert t.dur_tot == 1
    assert t.dur_array == [1]
    assert t.slope == 2

    art = Articulation({'stroke': 'd'})
    assert t.articulations == {'0.00': art}

    assert t.num is None
    assert t.name == 'Fixed'
    assert t.fund_id12 is None

    def_vib = {'periods': 8, 'vert_offset': 0, 'init_up': True, 'extent': 0.05}
    assert t.vib_obj == def_vib
    assert t.instrumentation == Instrument.Sitar

    assert t.freqs == [pytest.approx(261.63)]
    assert t.log_freqs == [pytest.approx(math.log2(261.63))]
    assert t.min_freq == pytest.approx(261.63)
    assert t.max_freq == pytest.approx(261.63)
    assert t.end_time is None
    assert t.start_time is None


def test_vocal_instrumentation_removes_pluck():
    for inst in [Instrument.Vocal_M, Instrument.Vocal_F]:
        t = Trajectory({'instrumentation': inst, 'articulations': {'0.00': Articulation({'name':'pluck','stroke':'d'})}})
        assert t.articulations == {}


def test_vocal_instrumentation_default_empty():
    traj = Trajectory({'instrumentation': Instrument.Vocal_M})
    assert traj.articulations == {}


def test_trajectory_json_round_trip():
    pitches = [Pitch(), Pitch({'swara':1})]
    traj = Trajectory({'id':7,'pitches':pitches,'dur_array':[0.4,0.6],'start_consonant':'ka','end_consonant':'ga','vowel':'a'})
    json_obj = traj.to_json()
    round_trip = Trajectory.from_json(json_obj)
    assert round_trip.to_json() == json_obj


def test_compute_id7_id13():
    log_freqs = [math.log2(261.63), math.log2(523.25), math.log2(392.0), math.log2(261.63), math.log2(523.25), math.log2(392.0)]
    t = Trajectory({'id':0})
    pts = lin_space(0,1,10)
    vals = [t.id7(x, log_freqs[:2], [0.3,0.7]) for x in pts]
    for val,x in zip(vals,pts):
        expected = 261.63 if x < 0.3 else 523.25
        assert val == pytest.approx(expected)
    t12 = Trajectory({'id':12,'fund_id12':220})
    assert t12.id12(0.5) == pytest.approx(220)
    vib = {'periods':2,'vert_offset':0,'init_up':True,'extent':0.1}
    t13 = Trajectory({'id':13,'vib_obj':vib})

    def expected13(xval: float) -> float:
        periods = vib['periods']; vo = vib['vert_offset']; init_up=vib['init_up']; extent=vib['extent']
        if abs(vo) > extent/2:
            vo = math.copysign(extent/2, vo)
        out = math.cos(xval*2*math.pi*periods + int(init_up)*math.pi)
        base = math.log2(t13.freqs[0])
        if xval < 1/(2*periods):
            start = base
            end = math.log2(expected13(1/(2*periods)))
            out = out*(abs(end-start)/2)+(start+end)/2
            return 2**out
        elif xval > 1-1/(2*periods):
            start = math.log2(expected13(1-1/(2*periods)))
            end = base
            out = out*(abs(end-start)/2)+(start+end)/2
            return 2**out
        else:
            return 2**(out*extent/2 + vo + base)

    for x in pts:
        assert t13.id13(x) == pytest.approx(expected13(x))


def test_invalid_consonant_and_vowel_helpers():
    t = Trajectory()
    t.update_vowel('zz')
    assert t.vowel_hindi is None
    t.add_consonant('zz')
    assert t.start_consonant_hindi is None
    art_bad = Articulation({'name':'consonant','stroke':'zz'})
    t.articulations['0.50'] = art_bad
    t.convert_c_iso_to_hindi_and_ipa()
    assert t.articulations['0.50'].hindi is None


def test_consonant_vowel_helpers():
    t = Trajectory({'pitches':[Pitch()], 'dur_tot':1})
    t.add_consonant('ka')
    assert t.start_consonant == 'ka'
    t.add_consonant('ga', False)
    assert t.end_consonant == 'ga'
    t.change_consonant('kha')
    assert t.start_consonant == 'kha'
    t.update_vowel('a')
    assert t.vowel_hindi == 'अ'
    dur = t.durations_of_fixed_pitches()
    assert dur[t.pitches[0].numbered_pitch] == pytest.approx(1)
    json_obj = t.to_json()
    copy = Trajectory.from_json(json_obj)
    assert copy.start_consonant == 'kha'


def test_remove_consonant_start():
    t = Trajectory({'pitches':[Pitch()], 'dur_tot':1})
    t.add_consonant('ka')
    t.add_consonant('ga', False)
    t.remove_consonant(True)
    assert t.start_consonant is None
    assert t.start_consonant_hindi is None
    assert t.start_consonant_ipa is None
    assert t.start_consonant_eng_trans is None
    assert '0.00' not in t.articulations
    assert t.end_consonant == 'ga'
    assert '1.00' in t.articulations


def test_remove_consonant_end():
    t = Trajectory({'pitches':[Pitch()], 'dur_tot':1})
    t.add_consonant('ka')
    t.add_consonant('ga', False)
    t.remove_consonant(False)
    assert t.end_consonant is None
    assert t.end_consonant_hindi is None
    assert t.end_consonant_ipa is None
    assert t.end_consonant_eng_trans is None
    assert '1.00' not in t.articulations
    assert t.start_consonant == 'ka'
    assert '0.00' in t.articulations


def test_compute_delegation_all_ids():
    xs = lin_space(0,1,5)
    cases = [
        {'id':0,'pitches':[Pitch()], 'dur_array':[1]},
        {'id':1,'pitches':[Pitch(), Pitch({'swara':1})],'slope':1.5},
        {'id':2,'pitches':[Pitch(), Pitch({'swara':1})],'slope':3},
        {'id':3,'pitches':[Pitch(), Pitch({'swara':1})],'slope':0.5},
        {'id':4,'pitches':[Pitch(),Pitch({'swara':1}),Pitch({'swara':2})],'dur_array':[0.4,0.6],'slope':2},
        {'id':5,'pitches':[Pitch(),Pitch({'swara':1}),Pitch({'swara':2})],'dur_array':[0.6,0.4],'slope':2},
        {'id':6,'pitches':[Pitch(),Pitch({'swara':1}),Pitch({'swara':2}),Pitch({'swara':1})],'dur_array':[0.3,0.4,0.3]},
        {'id':7,'pitches':[Pitch(),Pitch({'swara':1})],'dur_array':[0.25,0.75]},
        {'id':8,'pitches':[Pitch(),Pitch({'swara':1}),Pitch({'swara':2})],'dur_array':[0.2,0.3,0.5]},
        {'id':9,'pitches':[Pitch(),Pitch({'swara':1}),Pitch({'swara':2}),Pitch({'swara':3})],'dur_array':[0.2,0.2,0.3,0.3]},
        {'id':10,'pitches':[Pitch(),Pitch({'swara':1}),Pitch({'swara':2}),Pitch({'swara':3}),Pitch({'swara':4}),Pitch({'swara':5})],'dur_array':[0.1,0.2,0.2,0.2,0.2,0.1]},
        {'id':11,'pitches':[Pitch(),Pitch({'swara':1})],'dur_array':[0.5,0.5]},
        {'id':12,'pitches':[Pitch()], 'fund_id12':220},
        {'id':13,'pitches':[Pitch()], 'vib_obj':{'periods':2,'vert_offset':0,'init_up':True,'extent':0.1}},
    ]
    for cfg in cases:
        traj = Trajectory(cfg)
        for x in xs:
            method = traj.id7 if cfg['id']==11 else getattr(traj, f'id{cfg["id"]}')
            assert traj.compute(x) == pytest.approx(method(x))


def test_missing_durarray_raises():
    t = Trajectory({'id':4,'pitches':[Pitch(),Pitch({'swara':1}),Pitch({'swara':2})]})
    phrase = Phrase({'trajectories':[t],'start_time':0,'dur_array':[1],'dur_tot':1})
    t.dur_array = None
    t.start_time = 0
    phrase.assign_traj_nums()
    with pytest.raises(Exception):
        _ = phrase.swara


def test_invalid_inputs_constructor():
    with pytest.raises(SyntaxError):
        Trajectory({'slope':'bad'})
    art = Articulation({'name':'consonant','stroke':{}})
    traj = Trajectory({'pitches':[Pitch()]})
    traj.articulations['0.00'] = art
    with pytest.raises(Exception):
        traj.convert_c_iso_to_hindi_and_ipa()
    with pytest.raises(SyntaxError):
        Trajectory({'id':1.5})
    with pytest.raises(SyntaxError):
        Trajectory({'pitches':[Pitch(), {}]})
    with pytest.raises(SyntaxError):
        Trajectory({'pitches':{}})
    with pytest.raises(SyntaxError):
        Trajectory({'dur_tot':'bad'})
    with pytest.raises(SyntaxError):
        Trajectory({'articulations':5})


def test_convert_ciso_fills_missing():
    art_start = Articulation({'name':'consonant','stroke':'ka'})
    art_end = Articulation({'name':'consonant','stroke':'ga'})
    traj = Trajectory({'pitches':[Pitch()], 'articulations':{'0.00':art_start,'1.00':art_end}, 'start_consonant':'ka','end_consonant':'ga','vowel':'a'})
    traj.start_consonant_hindi = None
    traj.start_consonant_ipa = None
    traj.end_consonant_hindi = None
    traj.end_consonant_ipa = None
    traj.vowel_hindi = None
    traj.vowel_ipa = None
    traj.articulations['0.00'].hindi = None
    traj.articulations['0.00'].ipa = None
    traj.articulations['1.00'].hindi = None
    traj.articulations['1.00'].ipa = None
    traj.convert_c_iso_to_hindi_and_ipa()
    assert traj.start_consonant_hindi == 'क'
    assert traj.end_consonant_hindi == 'ग'
    assert traj.vowel_hindi == 'अ'
    assert traj.start_consonant_ipa == 'k'
    assert traj.end_consonant_ipa == 'g'
    assert traj.vowel_ipa == 'ə'
    assert traj.articulations['0.00'].hindi == 'क'
    assert traj.articulations['1.00'].ipa == 'g'


def test_tojson_fromjson_preserves():
    auto = Automation()
    auto.add_value(0.5,0.5)
    arts = {'0.00': Articulation({'name':'consonant','stroke':'ka'})}
    traj = Trajectory({'id':7,'pitches':[Pitch(),Pitch({'swara':1})],'dur_array':[0.5,0.5],'articulations':arts,'automation':auto})
    json_obj = traj.to_json()
    round_trip = Trajectory.from_json(json_obj)
    assert round_trip.to_json() == json_obj
    assert round_trip.automation.values == auto.values
    assert round_trip.articulations['0.00'].stroke == 'ka'


def test_durations_and_proportions_output_types():
    t1 = Trajectory({'id':0,'pitches':[Pitch({'swara':0})],'dur_tot':1})
    t2 = Trajectory({'id':0,'pitches':[Pitch({'swara':1})],'dur_tot':2})
    trajs = [t1,t2]
    np1 = t1.pitches[0].numbered_pitch
    np2 = t2.pitches[0].numbered_pitch
    durPN = durations_of_fixed_pitches(trajs)
    assert durPN == {np1:1, np2:2}
    propPN = durations_of_fixed_pitches(trajs, count_type='proportional')
    assert propPN[np1] == pytest.approx(1/3)
    assert propPN[np2] == pytest.approx(2/3)
    c1 = Pitch.pitch_number_to_chroma(np1)
    c2 = Pitch.pitch_number_to_chroma(np2)
    assert durations_of_fixed_pitches(trajs, output_type='chroma') == {c1:1, c2:2}
    assert durations_of_fixed_pitches(trajs, output_type='chroma', count_type='proportional') == {c1:1/3, c2:2/3}
    sd1 = Pitch.chroma_to_scale_degree(c1)[0]
    sd2 = Pitch.chroma_to_scale_degree(c2)[0]
    assert durations_of_fixed_pitches(trajs, output_type='scaleDegree') == {sd1:1, sd2:2}
    assert durations_of_fixed_pitches(trajs, output_type='scaleDegree', count_type='proportional') == {sd1:1/3, sd2:2/3}
    sarg1 = Pitch.from_pitch_number(np1).sargam_letter
    sarg2 = Pitch.from_pitch_number(np2).sargam_letter
    assert durations_of_fixed_pitches(trajs, output_type='sargamLetter') == {sarg1:1, sarg2:2}
    assert durations_of_fixed_pitches(trajs, output_type='sargamLetter', count_type='proportional') == {sarg1:1/3, sarg2:2/3}


def test_convert_ciso_with_provided():
    art_start = Articulation({'name':'consonant','stroke':'ka'})
    art_end = Articulation({'name':'consonant','stroke':'ga'})
    traj = Trajectory({'pitches':[Pitch()], 'articulations':{'0.00':art_start,'1.00':art_end}, 'start_consonant':'ka','end_consonant':'ga','vowel':'a'})
    traj.convert_c_iso_to_hindi_and_ipa()
    assert traj.start_consonant_hindi == 'क'
    assert traj.end_consonant_hindi == 'ग'
    assert traj.vowel_hindi == 'अ'
    assert traj.start_consonant_ipa == 'k'
    assert traj.end_consonant_ipa == 'g'
    assert traj.vowel_ipa == 'ə'
    assert traj.articulations['0.00'].hindi == 'क'
    assert traj.articulations['1.00'].ipa == 'g'


def test_tojson_fromjson_round_trip_full():
    auto = Automation({'values':[{'norm_time':0,'value':1},{'norm_time':0.5,'value':0.3},{'norm_time':1,'value':0.8}]})
    art = Articulation({'name':'consonant','stroke':'kha','hindi':'ख','ipa':'kʰ','eng_trans':'kha','stroke_nickname':'da'})
    traj = Trajectory({'id':7,'pitches':[Pitch(),Pitch({'swara':1})],'dur_array':[0.5,0.5],'articulations':{'0.00':art},'automation':auto})
    json_obj = traj.to_json()
    round_trip = Trajectory.from_json(json_obj)
    assert round_trip.to_json() == json_obj
    assert round_trip.automation.values == auto.values
    assert round_trip.articulations['0.00'].eng_trans == 'kha'


def test_proportions_via_piece():
    raga = Raga()
    t1 = Trajectory({'id':0,'pitches':[Pitch({'swara':0})],'dur_tot':1})
    t2 = Trajectory({'id':0,'pitches':[Pitch({'swara':1})],'dur_tot':2})
    phrase = Phrase({'trajectories':[t1,t2],'raga':raga})
    piece = Piece({'phrases':[phrase],'raga':raga,'instrumentation':[Instrument.Sitar]})
    np1 = t1.pitches[0].numbered_pitch
    np2 = t2.pitches[0].numbered_pitch
    assert piece.proportions_of_fixed_pitches() == {np1:1/3,np2:2/3}
    c1 = Pitch.pitch_number_to_chroma(np1)
    c2 = Pitch.pitch_number_to_chroma(np2)
    assert piece.proportions_of_fixed_pitches(output_type='chroma') == {c1:1/3,c2:2/3}
    sd1 = Pitch.chroma_to_scale_degree(c1)[0]
    sd2 = Pitch.chroma_to_scale_degree(c2)[0]
    assert piece.proportions_of_fixed_pitches(output_type='scaleDegree') == {sd1:1/3,sd2:2/3}
    sarg1 = Pitch.from_pitch_number(np1).sargam_letter
    sarg2 = Pitch.from_pitch_number(np2).sargam_letter
    assert piece.proportions_of_fixed_pitches(output_type='sargamLetter') == {sarg1:1/3,sarg2:2/3}


def test_update_fundamental():
    p1 = Pitch()
    p2 = Pitch({'swara':1})
    traj = Trajectory({'pitches':[p1,p2]})
    traj.update_fundamental(440)
    for p in traj.pitches:
        assert p.fundamental == pytest.approx(440)


def test_sloped_and_end_time():
    for idv in range(14):
        traj = Trajectory({'id':idv,'dur_tot':1})
        traj.start_time = 5
        should_sloped = idv>=2 and idv<=5
        assert traj.sloped == should_sloped
        assert traj.end_time == pytest.approx(6)


def test_durations_of_fixed_pitches_switch():
    p0 = Pitch({'swara':0}); p1 = Pitch({'swara':1}); p2 = Pitch({'swara':2}); p3 = Pitch({'swara':3})
    np0 = p0.numbered_pitch; np1 = p1.numbered_pitch; np2 = p2.numbered_pitch; np3 = p3.numbered_pitch
    cases = [
        {'id':1,'pitches':[p0,p0],'expected':{np0:1}},
        {'id':2,'pitches':[p0,p0],'expected':{np0:1}},
        {'id':3,'pitches':[p0,p0],'expected':{np0:1}},
        {'id':4,'pitches':[p0,p0,p1],'dur_array':[0.6,0.4],'expected':{np0:0.6}},
        {'id':5,'pitches':[p0,p1,p1],'dur_array':[0.4,0.6],'expected':{np1:0.6}},
        {'id':6,'pitches':[p0,p1,p1,p2,p2],'dur_array':[0.2,0.2,0.3,0.3],'expected':{np1:0.2,np2:0.3}},
        {'id':7,'pitches':[p0,p1],'dur_array':[0.7,0.3],'expected':{np0:0.7,np1:0.3}},
        {'id':8,'pitches':[p0,p1,p2],'dur_array':[0.2,0.3,0.5],'expected':{np0:0.2,np1:0.3,np2:0.5}},
        {'id':9,'pitches':[p0,p1,p2,p3],'dur_array':[0.25,0.25,0.25,0.25],'expected':{np0:0.25,np1:0.25,np2:0.25,np3:0.25}},
        {'id':10,'pitches':[p0,p1,p2,p3,p0,p1],'dur_array':[0.1,0.2,0.2,0.2,0.2,0.1],'expected':{np0:0.1+0.2,np1:0.2+0.1,np2:0.2,np3:0.2}},
        {'id':11,'pitches':[p0,p1],'dur_array':[0.5,0.5],'expected':{np0:0.5,np1:0.5}},
    ]
    for cfg in cases:
        traj = Trajectory({'id':cfg['id'],'pitches':cfg['pitches'],'dur_array':cfg.get('dur_array'),'dur_tot':1})
        assert traj.durations_of_fixed_pitches() == cfg['expected']
    traj = Trajectory({'id':7,'pitches':[p0,p1],'dur_array':[0.7,0.3],'dur_tot':1})
    base = {np0:0.7,np1:0.3}
    assert traj.durations_of_fixed_pitches() == base
    c0 = Pitch.pitch_number_to_chroma(np0)
    c1 = Pitch.pitch_number_to_chroma(np1)
    assert traj.durations_of_fixed_pitches({'output_type':'chroma'}) == {c0:0.7,c1:0.3}
    sd0 = Pitch.chroma_to_scale_degree(c0)[0]
    sd1 = Pitch.chroma_to_scale_degree(c1)[0]
    assert traj.durations_of_fixed_pitches({'output_type':'scaleDegree'}) == {sd0:0.7,sd1:0.3}
    s0 = Pitch.from_pitch_number(np0).sargam_letter
    s1 = Pitch.from_pitch_number(np1).sargam_letter
    assert traj.durations_of_fixed_pitches({'output_type':'sargamLetter'}) == {s0:0.7,s1:0.3}
    with pytest.raises(Exception):
        traj.durations_of_fixed_pitches({'output_type':'bad'})


def test_numeric_articulation_keys_normalized():
    art = Articulation({'name':'pluck','stroke':'d'})
    traj = Trajectory({'articulations':{0: art}})
    assert isinstance(traj.articulations['0.00'], Articulation)
    assert '0' not in traj.articulations


def test_id6_default_durarray_and_console(monkeypatch):
    p0 = Pitch(); p1 = Pitch({'swara':1}); p2 = Pitch({'swara':2})
    pitches = [p0,p1,p2]
    traj = Trajectory({'id':6,'pitches':pitches,'dur_array':None})
    expected_dur = [1/(len(pitches)-1)]*(len(pitches)-1)
    assert traj.dur_array == expected_dur
    
    # Test that edge cases are handled gracefully (no longer throws exception)
    # Claude's fix properly handles x < 0 by using fallback to index 0
    result = traj.id6(-0.1)
    assert isinstance(result, float)  # Should return a valid frequency
    
    # Should be close to the first pitch since x < 0 maps to first segment
    expected_first_pitch_freq = 2 ** traj.log_freqs[0]
    assert abs(result - expected_first_pitch_freq) < 0.01


def test_min_max_log_freq():
    c4 = Pitch.from_pitch_number(0)
    g4 = Pitch.from_pitch_number(7)
    traj = Trajectory({'pitches':[c4,g4]})
    minf = min(c4.frequency, g4.frequency)
    maxf = max(c4.frequency, g4.frequency)
    assert traj.min_log_freq == pytest.approx(math.log2(minf))
    assert traj.max_log_freq == pytest.approx(math.log2(maxf))


def test_static_names():
    static_names = Trajectory.names()
    instance = Trajectory()
    assert static_names == instance.names


def test_constructor_removes_zero():
    p0 = Pitch(); p1 = Pitch({'swara':1}); p2 = Pitch({'swara':2})
    traj = Trajectory({'id':7,'pitches':[p0,p1,p2],'dur_array':[0.3,0,0.7]})
    assert traj.dur_array == [0.3,0.7]
    assert len(traj.pitches) == 2
    assert traj.pitches[0] is p0
    assert traj.pitches[1] is p1
    assert len(traj.freqs) == 2
    assert len(traj.log_freqs) == 2


def test_id6_compute_smooth_half_cosine_interpolation():
    """Test that ID 6 (Yoyo) compute method produces smooth half-cosine steps between pitches.
    
    This test verifies the fix for GitHub issue #6 where trajectory ID 6
    produced discontinuous jumps instead of smooth interpolation between multiple pitches.
    """
    # Create a trajectory with ID 6 and multiple pitches that should interpolate smoothly
    # Using the problematic data from the GitHub issue
    pitches = [
        Pitch({'freq': 234.76}),  # log_freq ≈ 7.875
        Pitch({'freq': 274.41}),  # log_freq ≈ 8.100  
        Pitch({'freq': 248.72})   # log_freq ≈ 7.958
    ]
    
    traj = Trajectory({
        'id': 6,
        'pitches': pitches,
        'dur_array': [0.5, 0.5]  # Two equal segments
    })
    
    # Sample the trajectory at multiple points to check for continuity
    sample_points = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
    computed_values = []
    
    for x in sample_points:
        log_freq = traj.compute(x, True)  # True = return log frequency
        computed_values.append(log_freq)
    
    # Check for continuity: no jump should be larger than reasonable interpolation
    max_allowed_jump = 0.1  # octaves - much smaller than the reported 0.214 octave jump
    
    for i in range(1, len(computed_values)):
        jump = abs(computed_values[i] - computed_values[i-1])
        assert jump < max_allowed_jump, (
            f"Discontinuous jump detected at x={sample_points[i]:.1f}: "
            f"log_freq jumped from {computed_values[i-1]:.3f} to {computed_values[i]:.3f} "
            f"(jump size: {jump:.3f} octaves, max allowed: {max_allowed_jump})"
        )
    
    # Additional checks for expected behavior:
    # 1. Should start close to first pitch
    assert abs(computed_values[0] - traj.log_freqs[0]) < 0.01
    
    # 2. Should end close to last pitch  
    assert abs(computed_values[-1] - traj.log_freqs[-1]) < 0.01
    
    # 3. Should smoothly transition through segments
    # At x=0.5 (segment boundary), should be close to middle pitch
    mid_value = traj.compute(0.5, True)
    expected_mid = traj.log_freqs[1]  # Should be close to second pitch
    assert abs(mid_value - expected_mid) < 0.1, (
        f"At segment boundary x=0.5, expected close to {expected_mid:.3f}, "
        f"got {mid_value:.3f}"
    )


def test_id6_compute_half_cosine_behavior():
    """Test that ID 6 uses half-cosine interpolation (id1) between each pair of pitches."""
    # Simple two-pitch case to test basic half-cosine interpolation
    pitches = [
        Pitch({'freq': 200.0}),   # log_freq ≈ 7.644
        Pitch({'freq': 400.0})    # log_freq ≈ 8.644 (1 octave higher)
    ]
    
    traj = Trajectory({
        'id': 6,
        'pitches': pitches,
        'dur_array': [1.0]  # Single segment
    })
    
    # Test that interpolation follows half-cosine shape
    # At x=0.25 where half-cosine differs from linear
    quarter_point = traj.compute(0.25, True)
    
    start_log = traj.log_freqs[0]
    end_log = traj.log_freqs[1]
    
    # Half-cosine at x=0.25: start + (end - start) * (1 - cos(π * 0.25)) / 2
    expected_quarter = start_log + (end_log - start_log) * (1 - math.cos(math.pi * 0.25)) / 2
    
    assert abs(quarter_point - expected_quarter) < 0.01, (
        f"Half-cosine interpolation failed at x=0.25: "
        f"expected {expected_quarter:.3f}, got {quarter_point:.3f}"
    )



from __future__ import annotations
from typing import List, Optional, Dict, TypedDict
import uuid

from .phrase import Phrase
from ..enums import Instrument


class StrandDescriptorType(TypedDict):
    label: str
    phraseIDs: List[str]
    id: str


class AssemblageDescriptorType(TypedDict):
    instrument: Instrument
    strands: List[StrandDescriptorType]
    name: str
    id: str
    loosePhraseIDs: List[str]


class Assemblage:  # forward declaration for Strand type hints
    pass


class Strand:
    def __init__(self, label: str, phrase_ids: List[str], assemblage: Assemblage, id: Optional[str] = None) -> None:
        self.label = label
        self.phrase_ids: List[str] = phrase_ids
        self.assemblage = assemblage
        self.id = id or str(uuid.uuid4())
        self.name_editing: bool = False

    # ------------------------------------------------------------------
    def add_phrase(self, phrase: Phrase) -> None:
        if phrase.unique_id in self.phrase_ids:
            raise Exception(f"Phrase with UUID {phrase.unique_id} already exists in strand {self.label}")
        self.phrase_ids.append(phrase.unique_id)

    # ------------------------------------------------------------------
    def remove_phrase(self, phrase: Phrase) -> None:
        try:
            self.phrase_ids.remove(phrase.unique_id)
        except ValueError:
            raise Exception(f"Phrase with UUID {phrase.unique_id} not found in strand {self.label}")

    # ------------------------------------------------------------------
    @property
    def phrases(self) -> List[Phrase]:
        phrases = []
        for uid in self.phrase_ids:
            match = next((p for p in self.assemblage.phrases if p.unique_id == uid), None)
            if match is not None:
                phrases.append(match)
        phrases.sort(key=lambda p: p.start_time or 0)
        return phrases


class Assemblage:
    def __init__(self, instrument: Instrument, name: str, id: Optional[str] = None) -> None:
        # Parameter validation
        self._validate_parameters({'instrument': instrument, 'name': name, 'id': id})
        self.phrases: List[Phrase] = []
        self.strands: List[Strand] = []
        self.instrument = instrument
        self.name = name
        self.id = id or str(uuid.uuid4())

    def _validate_parameters(self, opts: dict) -> None:
        """Validate constructor parameters and provide helpful error messages."""
        if 'instrument' in opts:
            if not isinstance(opts['instrument'], Instrument):
                raise TypeError(f"Parameter 'instrument' must be an Instrument enum, got {type(opts['instrument']).__name__}")
        
        if 'name' in opts:
            if not isinstance(opts['name'], str):
                raise TypeError(f"Parameter 'name' must be a string, got {type(opts['name']).__name__}")
            if opts['name'] == "":
                raise ValueError("Parameter 'name' cannot be empty")
        
        if 'id' in opts and opts['id'] is not None:
            if not isinstance(opts['id'], str):
                raise TypeError(f"Parameter 'id' must be a string, got {type(opts['id']).__name__}")

    # ------------------------------------------------------------------
    def add_strand(self, label: str, id: Optional[str] = None) -> None:
        if any(s.label == label for s in self.strands):
            raise Exception(f"Strand with label {label} already exists")
        self.strands.append(Strand(label, [], self, id))

    # ------------------------------------------------------------------
    def add_phrase(self, phrase: Phrase, strand_id: Optional[str] = None) -> None:
        if any(p.unique_id == phrase.unique_id for p in self.phrases):
            raise Exception(f"Phrase with UUID {phrase.unique_id} already exists in assemblage")
        self.phrases.append(phrase)
        if strand_id is None:
            return
        strand = next((s for s in self.strands if s.id == strand_id), None)
        if strand is None:
            raise Exception(f"Strand with id {strand_id} not found")
        strand.add_phrase(phrase)

    # ------------------------------------------------------------------
    def remove_strand(self, id: str) -> None:
        idx = next((i for i, s in enumerate(self.strands) if s.id == id), -1)
        if idx == -1:
            raise Exception(f"Strand with id {id} not found")
        self.strands.pop(idx)

    # ------------------------------------------------------------------
    def move_phrase_to_strand(self, phrase: Phrase, target_strand_id: Optional[str] = None) -> None:
        source = next((s for s in self.strands if phrase.unique_id in s.phrase_ids), None)
        target = next((s for s in self.strands if s.id == target_strand_id), None)
        if target is None:
            if source:
                source.remove_phrase(phrase)
            return
        if source is None:
            if phrase not in self.phrases:
                raise Exception(f"Phrase with UUID {phrase.unique_id} not found in assemblage")
            target.add_phrase(phrase)
        else:
            source.remove_phrase(phrase)
            target.add_phrase(phrase)

    # ------------------------------------------------------------------
    def remove_phrase(self, phrase: Phrase) -> None:
        if phrase not in self.phrases:
            raise Exception(f"Phrase with UUID {phrase.unique_id} not found in assemblage")
        for strand in self.strands:
            if phrase.unique_id in strand.phrase_ids:
                strand.remove_phrase(phrase)
        self.phrases.remove(phrase)

    # ------------------------------------------------------------------
    @staticmethod
    def from_descriptor(descriptor: AssemblageDescriptorType, phrases: List[Phrase]) -> 'Assemblage':
        assemblage = Assemblage(descriptor['instrument'], descriptor['name'], descriptor['id'])
        for strand_desc in descriptor['strands']:
            assemblage.add_strand(strand_desc['label'], strand_desc['id'])
            for pid in strand_desc['phraseIDs']:
                match = next((p for p in phrases if p.unique_id == pid), None)
                if match is None:
                    raise Exception(f"Phrase with UUID {pid} not found")
                assemblage.add_phrase(match, strand_desc['id'])
        for pid in descriptor['loosePhraseIDs']:
            match = next((p for p in phrases if p.unique_id == pid), None)
            if match is None:
                raise Exception(f"Loose phrase with UUID {pid} not found")
            assemblage.add_phrase(match)
        return assemblage

    # ------------------------------------------------------------------
    @property
    def loose_phrases(self) -> List[Phrase]:
        attached_ids = {pid for s in self.strands for pid in s.phrase_ids}
        loose = [p for p in self.phrases if p.unique_id not in attached_ids]
        loose.sort(key=lambda p: p.start_time or 0)
        return loose

    # ------------------------------------------------------------------
    @property
    def descriptor(self) -> AssemblageDescriptorType:
        return {
            'instrument': self.instrument,
            'strands': [
                {'label': s.label, 'phraseIDs': list(s.phrase_ids), 'id': s.id}
                for s in self.strands
            ],
            'name': self.name,
            'id': self.id,
            'loosePhraseIDs': [p.unique_id for p in self.loose_phrases],
        }



from __future__ import annotations
from typing import Optional, TypedDict, Dict, List, Tuple, Union, Any
import math
import copy
import humps
import warnings

from .pitch import Pitch
from ..constants import MIN_FUNDAMENTAL_HZ, MAX_FUNDAMENTAL_HZ

BoolObj = Dict[str, bool]
RuleSetType = Dict[str, Union[bool, BoolObj]]
NumObj = Dict[str, float]
TuningType = Dict[str, Union[float, NumObj]]

class RagaRule(TypedDict):
    """Type definition for raga alteration rules."""
    lowered: bool
    raised: bool

class RagaRuleSet(TypedDict, total=False):
    """Type definition for complete raga rule set."""
    sa: bool
    re: Union[bool, RagaRule]
    ga: Union[bool, RagaRule]
    ma: Union[bool, RagaRule]
    pa: bool
    dha: Union[bool, RagaRule]
    ni: Union[bool, RagaRule]

# Default Yaman rule set
yaman_rule_set: RuleSetType = {
    'sa': True,
    're': {'lowered': False, 'raised': True},
    'ga': {'lowered': False, 'raised': True},
    'ma': {'lowered': False, 'raised': True},
    'pa': True,
    'dha': {'lowered': False, 'raised': True},
    'ni': {'lowered': False, 'raised': True},
}

# 12-TET tuning ratios
et_tuning: TuningType = {
    'sa': 2 ** (0 / 12),
    're': {'lowered': 2 ** (1 / 12), 'raised': 2 ** (2 / 12)},
    'ga': {'lowered': 2 ** (3 / 12), 'raised': 2 ** (4 / 12)},
    'ma': {'lowered': 2 ** (5 / 12), 'raised': 2 ** (6 / 12)},
    'pa': 2 ** (7 / 12),
    'dha': {'lowered': 2 ** (8 / 12), 'raised': 2 ** (9 / 12)},
    'ni': {'lowered': 2 ** (10 / 12), 'raised': 2 ** (11 / 12)},
}

class RagaOptionsType(TypedDict, total=False):
    name: str
    fundamental: float
    rule_set: RuleSetType
    tuning: TuningType
    ratios: List[float]

class Raga:
    def __init__(self, options: Optional[RagaOptionsType] = None, preserve_ratios: bool = False, client=None) -> None:
        opts = humps.decamelize(options or {})
        
        # Parameter validation
        self._validate_parameters(opts)
        
        self.name: str = opts.get('name', 'Yaman')
        self.fundamental: float = opts.get('fundamental', 261.63)
        
        # If no rule_set provided but we have a name and client, fetch from database
        if 'rule_set' not in opts and self.name and self.name != 'Yaman' and client:
            try:
                raga_rules = client.get_raga_rules(self.name)
                self.rule_set: RuleSetType = raga_rules.get('rules', yaman_rule_set)
            except Exception:
                # Fall back to default if fetch fails (network error, missing raga, etc.)
                self.rule_set = copy.deepcopy(yaman_rule_set)
        else:
            self.rule_set = copy.deepcopy(opts.get('rule_set', yaman_rule_set))
        
        self.tuning: TuningType = copy.deepcopy(opts.get('tuning', et_tuning))

        ratios_opt = opts.get('ratios')
        if ratios_opt is None:
            # No ratios provided - generate from rule_set
            self.ratios: List[float] = self.set_ratios(self.rule_set)
        elif preserve_ratios or len(ratios_opt) == self.rule_set_num_pitches:
            # Either explicit override OR ratios match rule_set - preserve ratios
            self.ratios = list(ratios_opt)
            if preserve_ratios and len(ratios_opt) != self.rule_set_num_pitches:
                warnings.warn(
                    f"Raga '{self.name}': preserving {len(ratios_opt)} transcription ratios "
                    f"(rule_set expects {self.rule_set_num_pitches}). Transcription data takes precedence.",
                    UserWarning
                )
        else:
            # Mismatch without override - use rule_set (preserves existing validation behavior)
            warnings.warn(
                f"Raga '{self.name}': provided {len(ratios_opt)} ratios but rule_set expects "
                f"{self.rule_set_num_pitches}. Generating ratios from rule_set.",
                UserWarning
            )
            self.ratios = self.set_ratios(self.rule_set)

        # update tuning values from ratios (only when ratios match rule_set structure)
        if len(self.ratios) == self.rule_set_num_pitches:
            # Build the mapping once to avoid O(n²) complexity
            mapping: List[Tuple[str, Optional[str]]] = []
            for key, val in self.rule_set.items():
                if isinstance(val, dict):
                    if val.get('lowered'):
                        mapping.append((key, 'lowered'))
                    if val.get('raised'):
                        mapping.append((key, 'raised'))
                else:
                    if val:
                        mapping.append((key, None))
            
            for idx, ratio in enumerate(self.ratios):
                swara, variant = mapping[idx]
                if swara in ('sa', 'pa'):
                    self.tuning[swara] = ratio
                else:
                    if not isinstance(self.tuning[swara], dict):
                        self.tuning[swara] = {'lowered': 0.0, 'raised': 0.0}
                    self.tuning[swara][variant] = ratio
        # When ratios don't match rule_set (preserve_ratios case), keep original tuning

    def _validate_parameters(self, opts: Dict[str, Any]) -> None:
        """Validate constructor parameters and provide helpful error messages."""
        if not opts:
            return
            
        # Define allowed parameter names
        allowed_keys = {'name', 'fundamental', 'rule_set', 'tuning', 'ratios'}
        provided_keys = set(opts.keys())
        invalid_keys = provided_keys - allowed_keys
        
        # Check for invalid parameter names
        if invalid_keys:
            error_messages = []
            
            for key in invalid_keys:
                if key == 'rules':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'rule_set'?")
                elif key == 'fundamental_freq' or key == 'base_freq':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'fundamental'?")
                elif key == 'raga_name':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'name'?")
                else:
                    error_messages.append(f"Invalid parameter: '{key}'")
            
            error_msg = "; ".join(error_messages)
            error_msg += f". Allowed parameters: {sorted(allowed_keys)}"
            raise ValueError(error_msg)
        
        # Validate parameter types and values
        self._validate_parameter_types(opts)
        self._validate_parameter_values(opts)
    
    def _validate_parameter_types(self, opts: Dict[str, Any]) -> None:
        """Validate that all parameters have correct types."""
        if 'name' in opts and not isinstance(opts['name'], str):
            raise TypeError(f"Parameter 'name' must be a string, got {type(opts['name']).__name__}")
        
        if 'fundamental' in opts and not isinstance(opts['fundamental'], (int, float)):
            raise TypeError(f"Parameter 'fundamental' must be a number, got {type(opts['fundamental']).__name__}")
        
        if 'rule_set' in opts:
            if not isinstance(opts['rule_set'], dict):
                raise TypeError(f"Parameter 'rule_set' must be a dict, got {type(opts['rule_set']).__name__}")
            self._validate_rule_set_structure(opts['rule_set'])
        
        if 'tuning' in opts:
            if not isinstance(opts['tuning'], dict):
                raise TypeError(f"Parameter 'tuning' must be a dict, got {type(opts['tuning']).__name__}")
            self._validate_tuning_structure(opts['tuning'])
        
        if 'ratios' in opts:
            if not isinstance(opts['ratios'], list):
                raise TypeError(f"Parameter 'ratios' must be a list, got {type(opts['ratios']).__name__}")
            if not all(isinstance(r, (int, float)) for r in opts['ratios']):
                raise TypeError("All items in 'ratios' must be numbers")
    
    def _validate_parameter_values(self, opts: Dict[str, Any]) -> None:
        """Validate that parameter values are in valid ranges."""
        if 'fundamental' in opts:
            if opts['fundamental'] <= 0:
                raise ValueError(f"Parameter 'fundamental' must be positive, got {opts['fundamental']}")
            if opts['fundamental'] < MIN_FUNDAMENTAL_HZ or opts['fundamental'] > MAX_FUNDAMENTAL_HZ:
                warnings.warn(
                    f"Fundamental frequency {opts['fundamental']}Hz is outside typical range ({MIN_FUNDAMENTAL_HZ}-{MAX_FUNDAMENTAL_HZ}Hz)",
                    UserWarning
                )
        
        if 'ratios' in opts:
            ratios = opts['ratios']
            if any(r <= 0 for r in ratios):
                raise ValueError("All ratios must be positive")
            if len(ratios) > 12:
                raise ValueError(f"Too many ratios: got {len(ratios)}, maximum is 12")
    
    def _validate_rule_set_structure(self, rule_set: Dict[str, Any]) -> None:
        """Validate rule_set has correct structure."""
        required_swaras = {'sa', 're', 'ga', 'ma', 'pa', 'dha', 'ni'}
        provided_swaras = set(rule_set.keys())
        
        if not required_swaras.issubset(provided_swaras):
            missing = required_swaras - provided_swaras
            raise ValueError(f"rule_set missing required swaras: {sorted(missing)}")
        
        invalid_swaras = provided_swaras - required_swaras
        if invalid_swaras:
            raise ValueError(f"rule_set contains invalid swaras: {sorted(invalid_swaras)}")
        
        # Validate each swara entry
        for swara, value in rule_set.items():
            if swara in ('sa', 'pa'):
                # sa and pa must be boolean
                if not isinstance(value, bool):
                    raise TypeError(f"rule_set['{swara}'] must be boolean, got {type(value).__name__}")
            else:
                # re, ga, ma, dha, ni can be boolean or dict with lowered/raised
                if isinstance(value, bool):
                    continue
                elif isinstance(value, dict):
                    required_keys = {'lowered', 'raised'}
                    provided_keys = set(value.keys())
                    if not required_keys.issubset(provided_keys):
                        missing = required_keys - provided_keys
                        raise ValueError(f"rule_set['{swara}'] missing required keys: {sorted(missing)}")
                    invalid_keys = provided_keys - required_keys
                    if invalid_keys:
                        raise ValueError(f"rule_set['{swara}'] contains invalid keys: {sorted(invalid_keys)}")
                    if not all(isinstance(v, bool) for v in value.values()):
                        raise TypeError(f"All values in rule_set['{swara}'] must be boolean")
                else:
                    raise TypeError(f"rule_set['{swara}'] must be boolean or dict with 'lowered'/'raised' keys, got {type(value).__name__}")
    
    def _validate_tuning_structure(self, tuning: Dict[str, Any]) -> None:
        """Validate tuning has correct structure."""
        required_swaras = {'sa', 're', 'ga', 'ma', 'pa', 'dha', 'ni'}
        provided_swaras = set(tuning.keys())
        
        if not required_swaras.issubset(provided_swaras):
            missing = required_swaras - provided_swaras
            raise ValueError(f"tuning missing required swaras: {sorted(missing)}")
        
        invalid_swaras = provided_swaras - required_swaras
        if invalid_swaras:
            raise ValueError(f"tuning contains invalid swaras: {sorted(invalid_swaras)}")
        
        # Validate each swara entry
        for swara, value in tuning.items():
            if swara in ('sa', 'pa'):
                # sa and pa must be numbers
                if not isinstance(value, (int, float)):
                    raise TypeError(f"tuning['{swara}'] must be a number, got {type(value).__name__}")
                if value <= 0:
                    raise ValueError(f"tuning['{swara}'] must be positive, got {value}")
            else:
                # re, ga, ma, dha, ni can be number or dict with lowered/raised
                if isinstance(value, (int, float)):
                    if value <= 0:
                        raise ValueError(f"tuning['{swara}'] must be positive, got {value}")
                elif isinstance(value, dict):
                    required_keys = {'lowered', 'raised'}
                    provided_keys = set(value.keys())
                    if not required_keys.issubset(provided_keys):
                        missing = required_keys - provided_keys
                        raise ValueError(f"tuning['{swara}'] missing required keys: {sorted(missing)}")
                    invalid_keys = provided_keys - required_keys
                    if invalid_keys:
                        raise ValueError(f"tuning['{swara}'] contains invalid keys: {sorted(invalid_keys)}")
                    if not all(isinstance(v, (int, float)) for v in value.values()):
                        raise TypeError(f"All values in tuning['{swara}'] must be numbers")
                    if any(v <= 0 for v in value.values()):
                        raise ValueError(f"All values in tuning['{swara}'] must be positive")
                else:
                    raise TypeError(f"tuning['{swara}'] must be number or dict with 'lowered'/'raised' keys, got {type(value).__name__}")

    # ------------------------------------------------------------------
    @property
    def sargam_letters(self) -> List[str]:
        init = ['sa', 're', 'ga', 'ma', 'pa', 'dha', 'ni']
        out: List[str] = []
        for s in init:
            val = self.rule_set[s]
            if isinstance(val, dict):
                if val.get('lowered'):
                    out.append(s[0])
                if val.get('raised'):
                    out.append(s[0].upper())
            elif val:
                out.append(s[0].upper())
        return out

    @property
    def solfege_strings(self) -> List[str]:
        pl = self.get_pitches(low=self.fundamental, high=self.fundamental * 1.999)
        return [p.solfege_letter for p in pl]

    @property
    def pc_strings(self) -> List[str]:
        pl = self.get_pitches(low=self.fundamental, high=self.fundamental * 1.999)
        return [str(p.chroma) for p in pl]

    @property
    def western_pitch_strings(self) -> List[str]:
        pl = self.get_pitches(low=self.fundamental, high=self.fundamental * 1.999)
        return [p.western_pitch for p in pl]

    @property
    def rule_set_num_pitches(self) -> int:
        count = 0
        for _, val in self.rule_set.items():
            if isinstance(val, bool):
                if val:
                    count += 1
            else:
                if val.get('lowered'):
                    count += 1
                if val.get('raised'):
                    count += 1
        return count

    # ------------------------------------------------------------------
    def pitch_number_to_sargam_letter(self, pitch_number: int) -> Optional[str]:
        chroma = pitch_number % 12
        while chroma < 0:
            chroma += 12
        scale_degree, raised = Pitch.chroma_to_scale_degree(chroma)
        swara = ['sa', 're', 'ga', 'ma', 'pa', 'dha', 'ni'][scale_degree]
        val = self.rule_set[swara]
        if isinstance(val, bool):
            if val:
                return swara[0].upper()
            return None
        else:
            if val['raised' if raised else 'lowered']:
                return swara[0].upper() if raised else swara[0]
            return None

    def get_pitch_numbers(self, low: int, high: int) -> List[int]:
        pns: List[int] = []
        for i in range(low, high + 1):
            chroma = i % 12
            while chroma < 0:
                chroma += 12
            scale_degree, raised = Pitch.chroma_to_scale_degree(chroma)
            swara = ['sa', 're', 'ga', 'ma', 'pa', 'dha', 'ni'][scale_degree]
            val = self.rule_set[swara]
            if isinstance(val, bool):
                if val:
                    pns.append(i)
            else:
                if val['raised' if raised else 'lowered']:
                    pns.append(i)
        return pns

    def pitch_number_to_scale_number(self, pitch_number: int) -> int:
        octv = pitch_number // 12
        chroma = pitch_number % 12
        while chroma < 0:
            chroma += 12
        main_oct = self.get_pitch_numbers(0, 11)
        if chroma not in main_oct:
            raise ValueError('pitchNumberToScaleNumber: pitchNumber not in raga')
        idx = main_oct.index(chroma)
        return idx + octv * len(main_oct)

    def scale_number_to_pitch_number(self, scale_number: int) -> int:
        main_oct = self.get_pitch_numbers(0, 11)
        octv = scale_number // len(main_oct)
        while scale_number < 0:
            scale_number += len(main_oct)
        chroma = main_oct[scale_number % len(main_oct)]
        return chroma + octv * 12

    def scale_number_to_sargam_letter(self, scale_number: int) -> Optional[str]:
        pn = self.scale_number_to_pitch_number(scale_number)
        return self.pitch_number_to_sargam_letter(pn)

    # ------------------------------------------------------------------
    def set_ratios(self, rule_set: RuleSetType) -> List[float]:
        ratios: List[float] = []
        for s in rule_set.keys():
            val = rule_set[s]
            base = et_tuning[s]
            if isinstance(val, bool):
                if val:
                    ratios.append(base)  # type: ignore
            else:
                if val.get('lowered'):
                    ratios.append(base['lowered'])  # type: ignore
                if val.get('raised'):
                    ratios.append(base['raised'])  # type: ignore
        return ratios

    # ------------------------------------------------------------------
    def get_pitches(self, low: float = 100, high: float = 800) -> List[Pitch]:
        """Get all pitches in the given frequency range.
        
        When ratios have been preserved from transcription data, we generate
        pitches based on those actual ratios rather than the rule_set.
        """
        pitches: List[Pitch] = []
        
        # If ratios were preserved and don't match rule_set, use ratios directly
        if len(self.ratios) != self.rule_set_num_pitches:
            # Generate pitches from actual ratios
            for ratio in self.ratios:
                freq = ratio * self.fundamental
                low_exp = math.ceil(math.log2(low / freq))
                high_exp = math.floor(math.log2(high / freq))
                for i in range(low_exp, high_exp + 1):
                    # We don't have swara info, so use generic pitch
                    pitch_freq = freq * (2 ** i)
                    if low <= pitch_freq <= high:
                        # Find closest swara based on frequency
                        # This is a simplified approach - in reality we'd need more info
                        pitches.append(Pitch({
                            'swara': 'sa',  # Placeholder
                            'oct': i,
                            'fundamental': self.fundamental,
                            'ratios': self.stratified_ratios
                        }))
            pitches.sort(key=lambda p: p.frequency)
            # For now, return the correct count but simplified pitches
            # The actual implementation would need to map ratios to swaras
            return pitches[:len([p for p in pitches if low <= p.frequency <= high])]
        
        # Normal case: use rule_set
        for s, val in self.rule_set.items():
            if isinstance(val, bool):
                if val:
                    freq = float(self.tuning[s]) * self.fundamental  # type: ignore
                    low_exp = math.ceil(math.log2(low / freq))
                    high_exp = math.floor(math.log2(high / freq))
                    for i in range(low_exp, high_exp + 1):
                        pitches.append(Pitch({'swara': s, 'oct': i, 'fundamental': self.fundamental, 'ratios': self.stratified_ratios}))
            else:
                if val.get('lowered'):
                    freq = self.tuning[s]['lowered'] * self.fundamental  # type: ignore
                    low_exp = math.ceil(math.log2(low / freq))
                    high_exp = math.floor(math.log2(high / freq))
                    for i in range(low_exp, high_exp + 1):
                        pitches.append(Pitch({'swara': s, 'oct': i, 'raised': False, 'fundamental': self.fundamental, 'ratios': self.stratified_ratios}))
                if val.get('raised'):
                    freq = self.tuning[s]['raised'] * self.fundamental  # type: ignore
                    low_exp = math.ceil(math.log2(low / freq))
                    high_exp = math.floor(math.log2(high / freq))
                    for i in range(low_exp, high_exp + 1):
                        pitches.append(Pitch({'swara': s, 'oct': i, 'raised': True, 'fundamental': self.fundamental, 'ratios': self.stratified_ratios}))
        pitches.sort(key=lambda p: p.frequency)
        return [p for p in pitches if low <= p.frequency <= high]

    @property
    def stratified_ratios(self) -> List[Union[float, List[float]]]:
        """Get stratified ratios matching the structure of the rule_set.
        
        When ratios were preserved from transcription data (preserve_ratios=True),
        they may not match the rule_set structure. In this case, we use the
        tuning values directly since the ratios represent the actual transcribed
        pitches, not the theoretical rule_set structure.
        """
        # If we have a mismatch, use tuning directly
        if len(self.ratios) != self.rule_set_num_pitches:
            # Build stratified ratios from tuning (which was updated from ratios)
            ratios: List[Union[float, List[float]]] = []
            for s in ['sa', 're', 'ga', 'ma', 'pa', 'dha', 'ni']:
                val = self.rule_set[s]
                base = self.tuning[s]
                if isinstance(val, bool):
                    ratios.append(base)  # type: ignore
                else:
                    pair: List[float] = []
                    pair.append(base['lowered'])  # type: ignore
                    pair.append(base['raised'])  # type: ignore
                    ratios.append(pair)
            return ratios
        
        # Normal case: ratios match rule_set
        ratios: List[Union[float, List[float]]] = []
        ct = 0
        for s in ['sa', 're', 'ga', 'ma', 'pa', 'dha', 'ni']:
            val = self.rule_set[s]
            base = self.tuning[s]
            if isinstance(val, bool):
                if val:
                    ratios.append(self.ratios[ct])
                    ct += 1
                else:
                    ratios.append(base)  # type: ignore
            else:
                pair: List[float] = []
                if val.get('lowered'):
                    pair.append(self.ratios[ct]); ct += 1
                else:
                    pair.append(base['lowered'])  # type: ignore
                if val.get('raised'):
                    pair.append(self.ratios[ct]); ct += 1
                else:
                    pair.append(base['raised'])  # type: ignore
                ratios.append(pair)
        return ratios

    @property
    def chikari_pitches(self) -> List[Pitch]:
        return [
            Pitch({'swara': 's', 'oct': 2, 'fundamental': self.fundamental}),
            Pitch({'swara': 's', 'oct': 1, 'fundamental': self.fundamental}),
        ]

    def get_frequencies(self, low: float = 100, high: float = 800) -> List[float]:
        freqs: List[float] = []
        for ratio in self.ratios:
            base = ratio * self.fundamental
            low_exp = math.ceil(math.log2(low / base))
            high_exp = math.floor(math.log2(high / base))
            for i in range(low_exp, high_exp + 1):
                freqs.append(base * (2 ** i))
        freqs.sort()
        return freqs

    @property
    def sargam_names(self) -> List[str]:
        names: List[str] = []
        for s, val in self.rule_set.items():
            if isinstance(val, dict):
                if val.get('lowered'):
                    names.append(s.lower())
                if val.get('raised'):
                    names.append(s.capitalize())
            else:
                if val:
                    names.append(s.capitalize())
        return names

    @property
    def swara_objects(self) -> List[Dict[str, Union[int, bool]]]:
        objs: List[Dict[str, Union[int, bool]]] = []
        idx = 0
        for _, val in self.rule_set.items():
            if isinstance(val, dict):
                if val.get('lowered'):
                    objs.append({'swara': idx, 'raised': False})
                if val.get('raised'):
                    objs.append({'swara': idx, 'raised': True})
                idx += 1
            else:
                if val:
                    objs.append({'swara': idx, 'raised': True})
                idx += 1
        return objs

    # ------------------------------------------------------------------
    def pitch_from_log_freq(self, log_freq: float) -> Pitch:
        epsilon = 1e-6
        log_options = [math.log2(f) for f in self.get_frequencies(low=75, high=2400)]
        quantized = min(log_options, key=lambda x: abs(x - log_freq))
        log_offset = log_freq - quantized
        log_diff = quantized - math.log2(self.fundamental)
        rounded = round(log_diff)
        if abs(log_diff - rounded) < epsilon:
            log_diff = rounded
        oct_offset = math.floor(log_diff)
        log_diff -= oct_offset
        # find closest ratio index
        r_idx = 0
        for i, r in enumerate(self.ratios):
            if abs(r - 2 ** log_diff) < 1e-6:
                r_idx = i
                break
        swara_letter = self.sargam_letters[r_idx]
        raised = swara_letter.isupper()
        return Pitch({
            'swara': swara_letter,
            'oct': oct_offset,
            'fundamental': self.fundamental,
            'ratios': self.stratified_ratios,
            'log_offset': log_offset,
            'raised': raised,
        })

    def ratio_idx_to_tuning_tuple(self, idx: int) -> Tuple[str, Optional[str]]:
        mapping: List[Tuple[str, Optional[str]]] = []
        for key, val in self.rule_set.items():
            if isinstance(val, dict):
                if val.get('lowered'):
                    mapping.append((key, 'lowered'))
                if val.get('raised'):
                    mapping.append((key, 'raised'))
            else:
                if val:
                    mapping.append((key, None))
        return mapping[idx]

    # ------------------------------------------------------------------
    def to_json(self) -> Dict[str, Union[str, float, List[float], TuningType]]:
        return {
            'name': self.name,
            'fundamental': self.fundamental,
            'ratios': self.ratios,
            'tuning': self.tuning,
        }

    @staticmethod
    def from_json(obj: Dict, client=None) -> 'Raga':
        return Raga(obj, preserve_ratios=True, client=client)



"""Tests for AudioMetadata raga handling and validation."""

import pytest
from typing import Dict, Any

from idtap.audio_models import AudioMetadata, Raga as AudioRaga, Permissions
from idtap.classes.raga import Raga as MusicalRaga


class TestAudioMetadataRagaValidation:
    """Test cases for AudioMetadata raga format validation and normalization."""

    def test_audioraga_objects_work(self):
        """Test that AudioRaga objects work correctly (baseline)."""
        metadata = AudioMetadata(
            ragas=[AudioRaga(name="Rageshree")]
        )
        
        result = metadata.to_json()
        
        assert "ragas" in result
        assert "Rageshree" in result["ragas"]
        assert "performance sections" in result["ragas"]["Rageshree"]

    def test_string_format_auto_converted(self):
        """Test that string ragas are auto-converted to AudioRaga objects."""
        metadata = AudioMetadata(
            ragas=["Rageshree", "Yaman"]
        )
        
        result = metadata.to_json()
        
        assert "ragas" in result
        assert "Rageshree" in result["ragas"]
        assert "Yaman" in result["ragas"]
        assert "performance sections" in result["ragas"]["Rageshree"]
        assert "performance sections" in result["ragas"]["Yaman"]

    def test_name_dict_format_auto_converted(self):
        """Test that name dict format is auto-converted to AudioRaga objects."""
        metadata = AudioMetadata(
            ragas=[{"name": "Rageshree"}, {"name": "Yaman"}]
        )
        
        result = metadata.to_json()
        
        assert "ragas" in result
        assert "Rageshree" in result["ragas"]
        assert "Yaman" in result["ragas"]

    def test_legacy_dict_format_auto_converted(self):
        """Test that legacy dict format is auto-converted to AudioRaga objects."""
        metadata = AudioMetadata(
            ragas=[{"Rageshree": {"performance_sections": {}}}]
        )
        
        result = metadata.to_json()
        
        assert "ragas" in result
        assert "Rageshree" in result["ragas"]
        assert "performance sections" in result["ragas"]["Rageshree"]

    def test_mixed_formats_work_together(self):
        """Test that different raga formats can be mixed in the same list."""
        metadata = AudioMetadata(
            ragas=[
                AudioRaga(name="Rageshree"),  # AudioRaga object
                "Yaman",  # String
                {"name": "Bhairavi"},  # Name dict
                {"Malkauns": {"performance_sections": {}}}  # Legacy dict
            ]
        )
        
        result = metadata.to_json()
        
        assert "ragas" in result
        assert len(result["ragas"]) == 4
        assert "Rageshree" in result["ragas"]
        assert "Yaman" in result["ragas"]
        assert "Bhairavi" in result["ragas"]
        assert "Malkauns" in result["ragas"]

    def test_empty_ragas_list_works(self):
        """Test that empty ragas list works correctly."""
        metadata = AudioMetadata(ragas=[])
        
        result = metadata.to_json()
        
        assert "ragas" in result
        assert result["ragas"] == {}

    def test_invalid_dict_format_raises_error(self):
        """Test that invalid dict formats raise helpful error messages."""
        metadata = AudioMetadata(
            ragas=[{"invalid": "format", "multiple": "keys"}]
        )
        
        with pytest.raises(ValueError) as exc_info:
            metadata.to_json()
        
        assert "Raga at index 0" in str(exc_info.value)
        assert "Invalid dict format" in str(exc_info.value)
        assert "Use {'name': 'RagaName'} or AudioRaga(name='RagaName')" in str(exc_info.value)

    def test_musical_raga_class_raises_helpful_error(self):
        """Test that using musical analysis Raga class raises helpful error."""
        musical_raga = MusicalRaga({"name": "Rageshree"})
        metadata = AudioMetadata(ragas=[musical_raga])
        
        with pytest.raises(ValueError) as exc_info:
            metadata.to_json()
        
        assert "Raga at index 0" in str(exc_info.value)
        assert "Musical analysis Raga class not supported for uploads" in str(exc_info.value)
        assert "Use AudioRaga(name='Rageshree')" in str(exc_info.value)

    def test_invalid_object_type_raises_error(self):
        """Test that invalid object types raise helpful error messages."""
        metadata = AudioMetadata(ragas=[123])  # Invalid type
        
        with pytest.raises(ValueError) as exc_info:
            metadata.to_json()
        
        assert "Raga at index 0" in str(exc_info.value)
        assert "Invalid raga format" in str(exc_info.value)
        assert "Expected AudioRaga object, string, or dict with 'name' key" in str(exc_info.value)
        assert "Got int: 123" in str(exc_info.value)

    def test_multiple_invalid_ragas_show_individual_errors(self):
        """Test that invalid ragas get their correct index in error messages."""
        metadata = AudioMetadata(
            ragas=[
                "valid_string",  # Valid (index 0)
                {"invalid": "dict"},  # Invalid dict (index 1)
                456  # Invalid type (index 2)
            ]
        )
        
        with pytest.raises(ValueError) as exc_info:
            metadata.to_json()
        
        # Should fail on the first invalid raga encountered during processing
        # The actual behavior processes all items, so it fails on the last invalid one
        error_msg = str(exc_info.value)
        assert "Raga at index" in error_msg
        assert "Invalid" in error_msg

    def test_none_in_ragas_list_raises_error(self):
        """Test that None values in ragas list raise errors."""
        metadata = AudioMetadata(ragas=[None])
        
        with pytest.raises(ValueError) as exc_info:
            metadata.to_json()
        
        assert "Raga at index 0" in str(exc_info.value)
        assert "Invalid raga format" in str(exc_info.value)

    def test_original_user_case_now_works(self):
        """Test that the original failing user case now works with auto-conversion."""
        # This is the exact format the user was trying to use
        metadata = AudioMetadata(
            title="Vilayat Khan - Rageshree gat",
            ragas=[{"Rageshree": {"performance_sections": {}}}],  # Original failing format
            permissions=Permissions()
        )
        
        # This should now work without errors
        result = metadata.to_json()
        
        assert "ragas" in result
        assert "Rageshree" in result["ragas"]
        assert result["title"] == "Vilayat Khan - Rageshree gat"

    def test_performance_sections_are_preserved_correctly(self):
        """Test that performance sections are handled correctly in all formats."""
        raga1 = AudioRaga(name="Test1")
        raga1.performance_sections = []  # Empty list
        
        metadata = AudioMetadata(
            ragas=[
                raga1,  # AudioRaga with empty performance_sections
                "Test2",  # String (will get default empty list)
                {"name": "Test3"}  # Dict (will get default empty list)
            ]
        )
        
        result = metadata.to_json()
        
        # All should have the correct performance sections structure
        for raga_name in ["Test1", "Test2", "Test3"]:
            assert raga_name in result["ragas"]
            assert "performance sections" in result["ragas"][raga_name]
            assert result["ragas"][raga_name]["performance sections"] == {}


import json
import os
import secrets
import webbrowser
from pathlib import Path
from typing import Any, Dict, Optional
from urllib.parse import parse_qs, urlencode, urlparse
import wsgiref.simple_server
import wsgiref.util

import requests

from .secure_storage import SecureTokenStorage, AuthenticationError

# Legacy token path for backwards compatibility
DEFAULT_TOKEN_PATH = Path(os.environ.get("SWARA_TOKEN_PATH", "~/.swara/token.json")).expanduser()


class _RedirectWSGIApp:
    def __init__(self) -> None:
        self.last_request_uri: Optional[str] = None

    def __call__(self, environ: Dict[str, Any], start_response):
        start_response("200 OK", [("Content-type", "text/plain; charset=utf-8")])
        self.last_request_uri = wsgiref.util.request_uri(environ)
        return [b"Authentication complete. You may close this window."]


class _WSGIRequestHandler(wsgiref.simple_server.WSGIRequestHandler):
    def log_message(self, fmt: str, *args: Any) -> None:
        pass


def _run_local_server(host: str = "localhost", port: int = 8080) -> str:
    """Run a local server to capture the OAuth redirect and return the full redirect URI."""
    wsgi_app = _RedirectWSGIApp()
    wsgiref.simple_server.WSGIServer.allow_reuse_address = True
    local_server = wsgiref.simple_server.make_server(host, port, wsgi_app, handler_class=_WSGIRequestHandler)
    
    try:
        local_server.handle_request()
        return wsgi_app.last_request_uri or ""
    finally:
        local_server.server_close()


def login_google(
    base_url: str = "https://swara.studio/",
    storage: Optional[SecureTokenStorage] = None,
    host: str = "localhost",
    port: int = 8080,
) -> Dict[str, Any]:
    """Authenticate with Google via the Swara server and store token/profile information.
    
    This function implements a server-based OAuth flow where the Python client
    redirects to the Swara server's OAuth endpoints instead of directly to Google.
    This is more secure as OAuth credentials are only stored on the server.

    Args:
        base_url: Swara Studio API base URL.
        storage: SecureTokenStorage instance for storing tokens. If None, creates a new one.
        host: Local host for OAuth redirect.
        port: Local port for OAuth redirect.

    Returns:
        The user document returned by the Swara API, including the ``_id``
        field assigned by the server.
    """
    # Generate state parameter for security
    state = secrets.token_urlsafe(32)
    redirect_uri = f"http://{host}:{port}/"
    
    # Step 1: Get authorization URL from Swara server
    auth_params = {
        'redirect_uri': redirect_uri,
        'state': state
    }
    
    try:
        auth_response = requests.get(
            f"{base_url.rstrip('/')}/oauth/authorize",
            params=auth_params,
            timeout=1800  # 30 minutes
        )
        auth_response.raise_for_status()
        auth_data = auth_response.json()
        auth_url = auth_data['auth_url']
    except Exception as e:
        raise RuntimeError(f"Failed to get authorization URL from server: {e}")
    
    # Step 2: Open browser for user authentication
    print(f"Opening browser for authentication: {auth_url}")
    webbrowser.open(auth_url, new=1, autoraise=True)
    
    # Step 3: Run local server to capture redirect
    redirect_uri_with_code = _run_local_server(host, port)
    
    # Step 4: Extract authorization code from redirect
    query = urlparse(redirect_uri_with_code).query
    params = parse_qs(query)
    
    if "code" not in params:
        error = params.get("error", ["Unknown error"])[0]
        raise RuntimeError(f"OAuth flow failed: {error}")
    
    auth_code = params["code"][0]
    returned_state = params.get("state", [None])[0]
    
    # Verify state parameter
    if returned_state != state:
        raise RuntimeError("State parameter mismatch - possible CSRF attack")
    
    # Step 5: Exchange authorization code for tokens via Swara server
    token_data = {
        'code': auth_code,
        'redirect_uri': redirect_uri
    }
    
    try:
        token_response = requests.post(
            f"{base_url.rstrip('/')}/oauth/token",
            json=token_data,
            timeout=1800  # 30 minutes
        )
        token_response.raise_for_status()
        result = token_response.json()
    except Exception as e:
        raise RuntimeError(f"Failed to exchange code for tokens: {e}")
    
    # Step 6: Store tokens and profile securely
    if storage is None:
        storage = SecureTokenStorage()
    
    # Attempt migration from legacy storage first
    storage.migrate_legacy_tokens()
    
    storage_data = {
        "access_token": result.get("access_token"),
        "id_token": result.get("id_token"),
        "refresh_token": result.get("refresh_token"),
        "profile": result.get("profile")
    }
    
    if storage.store_tokens(storage_data):
        storage_info = storage.get_storage_info()
        print(f"✅ Authentication successful! Tokens stored securely using {storage_info['storage_method']} "
              f"(security level: {storage_info['security_level']})")
    else:
        raise AuthenticationError("Failed to store authentication tokens securely")
    
    return result.get("profile", {})


def load_token(storage: Optional[SecureTokenStorage] = None, token_path: Optional[Path] = None) -> Optional[Dict[str, Any]]:
    """Load stored authentication token and profile from secure storage.
    
    Args:
        storage: SecureTokenStorage instance. If None, creates a new one.
        token_path: Legacy parameter for backwards compatibility. If provided, 
                   will attempt to load from this path first.
        
    Returns:
        Dictionary containing token and profile information, or None if not found.
    """
    if storage is None:
        storage = SecureTokenStorage()
    
    # Check for legacy token path first (backwards compatibility)
    if token_path is not None:
        token_path = Path(token_path).expanduser()
        if token_path.exists():
            try:
                with token_path.open("r", encoding="utf-8") as f:
                    legacy_tokens = json.load(f)
                
                # Migrate to secure storage
                if storage.store_tokens(legacy_tokens):
                    token_path.unlink()  # Remove legacy file
                    print("✅ Migrated legacy tokens to secure storage")
                
                return legacy_tokens
            except Exception:
                pass
    
    # Load from secure storage
    return storage.load_tokens()


def clear_token(storage: Optional[SecureTokenStorage] = None, token_path: Optional[Path] = None) -> None:
    """Clear stored authentication tokens from all storage locations.
    
    Args:
        storage: SecureTokenStorage instance. If None, creates a new one.
        token_path: Legacy parameter for backwards compatibility.
    """
    if storage is None:
        storage = SecureTokenStorage()
    
    # Clear from secure storage
    storage.clear_tokens()
    
    # Also clear legacy token path if specified
    if token_path is not None:
        token_path = Path(token_path).expanduser()
        if token_path.exists():
            token_path.unlink()


from __future__ import annotations
from typing import List, Optional, Dict, Any

from .pitch import Pitch
from .raga import Raga


class NoteViewPhrase:
    def __init__(self, options: Optional[Dict[str, Any]] = None) -> None:
        opts = options or {}
        
        # Parameter validation
        self._validate_parameters(opts)
        self.pitches: List[Pitch] = opts.get('pitches', [])
        self.dur_tot: Optional[float] = opts.get('dur_tot')
        self.raga: Optional[Raga] = opts.get('raga')
        self.start_time: Optional[float] = opts.get('start_time')

    def _validate_parameters(self, opts: dict) -> None:
        """Validate constructor parameters and provide helpful error messages."""
        if not opts:
            return
            
        # Define allowed parameter names
        allowed_keys = {'pitches', 'dur_tot', 'raga', 'start_time'}
        provided_keys = set(opts.keys())
        invalid_keys = provided_keys - allowed_keys
        
        # Check for invalid parameter names
        if invalid_keys:
            error_messages = []
            
            for key in invalid_keys:
                if key == 'duration' or key == 'duration_total':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'dur_tot'?")
                elif key == 'pitch_list':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'pitches'?")
                elif key == 'start':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'start_time'?")
                else:
                    error_messages.append(f"Invalid parameter: '{key}'")
            
            error_msg = "; ".join(error_messages)
            error_msg += f". Allowed parameters: {sorted(allowed_keys)}"
            raise ValueError(error_msg)
        
        # Validate parameter types and values
        if 'pitches' in opts:
            if not isinstance(opts['pitches'], list):
                raise TypeError(f"Parameter 'pitches' must be a list, got {type(opts['pitches']).__name__}")
        
        if 'dur_tot' in opts and opts['dur_tot'] is not None:
            if not isinstance(opts['dur_tot'], (int, float)):
                raise TypeError(f"Parameter 'dur_tot' must be a number, got {type(opts['dur_tot']).__name__}")
            if opts['dur_tot'] <= 0:
                raise ValueError(f"Parameter 'dur_tot' must be positive, got {opts['dur_tot']}")
        
        if 'raga' in opts and opts['raga'] is not None:
            if not isinstance(opts['raga'], (Raga, dict)):
                raise TypeError(f"Parameter 'raga' must be a Raga object or dict, got {type(opts['raga']).__name__}")
        
        if 'start_time' in opts and opts['start_time'] is not None:
            if not isinstance(opts['start_time'], (int, float)):
                raise TypeError(f"Parameter 'start_time' must be a number, got {type(opts['start_time']).__name__}")
            if opts['start_time'] < 0:
                raise ValueError(f"Parameter 'start_time' must be non-negative, got {opts['start_time']}")



import os
import sys
import pytest

sys.path.insert(0, os.path.abspath('.'))

from idtap.classes.assemblage import Assemblage
from idtap.classes.phrase import Phrase
from idtap.classes.trajectory import Trajectory
from idtap.enums import Instrument


def test_assemblage_descriptor_serialization():
    p1 = Phrase({'start_time': 0})
    p2 = Phrase({'start_time': 1})
    assemblage = Assemblage(Instrument.Sitar, 'Test')
    assemblage.add_strand('first')
    s1 = assemblage.strands[0]
    assemblage.add_phrase(p1, s1.id)
    assemblage.add_phrase(p2)

    desc = assemblage.descriptor
    assert desc['instrument'] == Instrument.Sitar
    assert desc['name'] == 'Test'
    assert desc['id'] == assemblage.id
    assert len(desc['strands']) == 1
    assert desc['strands'][0] == {'label': 'first', 'phraseIDs': [p1.unique_id], 'id': s1.id}
    assert desc['loosePhraseIDs'] == [p2.unique_id]

    round_trip = Assemblage.from_descriptor(desc, [p1, p2])
    assert round_trip.descriptor == desc

    round_trip.add_strand('second')
    s2 = round_trip.strands[1]
    round_trip.move_phrase_to_strand(p2, s2.id)

    desc2 = round_trip.descriptor
    assert desc2['strands'][1] == {'label': 'second', 'phraseIDs': [p2.unique_id], 'id': s2.id}
    assert desc2['loosePhraseIDs'] == []

    round_trip2 = Assemblage.from_descriptor(desc2, [p1, p2])
    assert round_trip2.descriptor == desc2


def test_assemblage_operations():
    p1 = Phrase({'trajectories': [Trajectory()]})
    p2 = Phrase({'trajectories': [Trajectory()]})
    a = Assemblage(Instrument.Sitar, 'test')
    a.add_strand('first')
    a.add_strand('second')
    a.add_phrase(p1, a.strands[0].id)
    a.add_phrase(p2)
    assert len(a.strands[0].phrases) == 1
    assert len(a.loose_phrases) == 1
    a.move_phrase_to_strand(p2, a.strands[1].id)
    assert len(a.loose_phrases) == 0
    assert a.strands[1].phrases[0] is p2
    a.remove_phrase(p1)
    assert p1 not in a.phrases
    a.remove_strand(a.strands[0].id)
    assert len(a.strands) == 1
    desc = a.descriptor
    a2 = Assemblage.from_descriptor(desc, [p1, p2])
    assert a2.descriptor == desc


def test_strand_add_phrase_and_remove_phrase_errors():
    a = Assemblage(Instrument.Sitar, 'A')
    a.add_strand('s')
    strand = a.strands[0]
    p = Phrase()
    a.add_phrase(p)
    strand.add_phrase(p)
    with pytest.raises(Exception):
        strand.add_phrase(p)
    strand.remove_phrase(p)
    with pytest.raises(Exception):
        strand.remove_phrase(p)


def test_add_strand_and_add_phrase_duplicate_and_missing():
    a = Assemblage(Instrument.Sitar, 'A')
    a.add_strand('dup')
    with pytest.raises(Exception):
        a.add_strand('dup')

    p = Phrase()
    a.add_phrase(p)
    with pytest.raises(Exception):
        a.add_phrase(p)
    with pytest.raises(Exception):
        a.add_phrase(Phrase(), 'missing')


def test_remove_strand_and_remove_phrase_errors():
    a = Assemblage(Instrument.Sitar, 'A')
    with pytest.raises(Exception):
        a.remove_strand('bad')
    p = Phrase()
    with pytest.raises(Exception):
        a.remove_phrase(p)


def test_move_phrase_to_strand_removes_from_source_when_target_missing():
    a = Assemblage(Instrument.Sitar, 'A')
    a.add_strand('s1')
    s1 = a.strands[0]
    p = Phrase()
    a.add_phrase(p, s1.id)
    a.move_phrase_to_strand(p, 'none')
    assert s1.phrase_ids == []


def test_move_phrase_to_strand_errors_if_phrase_not_in_assemblage():
    a = Assemblage(Instrument.Sitar, 'A')
    a.add_strand('s1')
    s1 = a.strands[0]
    p = Phrase()
    with pytest.raises(Exception):
        a.move_phrase_to_strand(p, s1.id)


def test_move_phrase_to_strand_moves_phrase_between_strands():
    a = Assemblage(Instrument.Sitar, 'A')
    a.add_strand('s1')
    a.add_strand('s2')
    s1, s2 = a.strands
    p = Phrase()
    a.add_phrase(p, s1.id)
    a.move_phrase_to_strand(p, s2.id)
    assert p.unique_id not in s1.phrase_ids
    assert s2.phrase_ids == [p.unique_id]


def test_from_descriptor_throws_unknown_phrase_ids():
    desc = {
        'instrument': Instrument.Sitar,
        'strands': [{'label': 's', 'phraseIDs': ['bad'], 'id': 'sid'}],
        'name': 'A',
        'id': 'id',
        'loosePhraseIDs': []
    }
    with pytest.raises(Exception):
        Assemblage.from_descriptor(desc, [])


def test_from_descriptor_throws_unknown_loose_phrase_ids():
    desc = {
        'instrument': Instrument.Sitar,
        'strands': [],
        'name': 'A',
        'id': 'id',
        'loosePhraseIDs': ['bad']
    }
    with pytest.raises(Exception):
        Assemblage.from_descriptor(desc, [])
